import copy
from unittest import mock

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.testing.compare import compare_images
import numpy as np
import pytest

from cabinetry.visualize import plot_model


def test_data_mc(tmp_path, caplog):
    fname = tmp_path / "fig.png"
    histo_dict_list = [
        {
            "label": "Background",
            "isData": False,
            "yields": np.asarray([12.5, 14]),
            "variable": "x",
        },
        {
            "label": "Signal",
            "isData": False,
            "yields": np.asarray([2, 5]),
            "variable": "x",
        },
        {
            "label": "Data",
            "isData": True,
            "yields": np.asarray([13, 15]),
            "variable": "x",
        },
    ]
    total_model_unc = np.sqrt([0.17, 0.29])
    bin_edges = np.asarray([1, 2, 3])

    fig = plot_model.data_mc(
        histo_dict_list,
        total_model_unc,
        bin_edges,
        figure_path=fname,
        label="Signal region\npre-fit",
    )
    assert (
        compare_images("tests/visualize/reference/data_mc.png", str(fname), 0) is None
    )

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert (
        compare_images("tests/visualize/reference/data_mc.png", str(fname), 0) is None
    )

    histo_dict_list_log = copy.deepcopy(histo_dict_list)
    histo_dict_list_log[0]["yields"] = np.asarray([2000, 14])
    histo_dict_list_log[2]["yields"] = np.asarray([2010, 15])
    total_model_unc_log = np.asarray([50, 1.5])
    bin_edges_log = np.asarray([10, 100, 1000])
    fname_log = fname.with_name(fname.stem + "_log" + fname.suffix)

    # automatic log scale
    plot_model.data_mc(
        histo_dict_list_log,
        total_model_unc_log,
        bin_edges_log,
        figure_path=fname,
        log_scale_x=True,
        label="Signal region\npre-fit",
    )
    assert (
        compare_images("tests/visualize/reference/data_mc_log.png", str(fname_log), 0)
        is None
    )

    # linear scale forced
    plot_model.data_mc(
        histo_dict_list,
        total_model_unc,
        bin_edges,
        figure_path=fname,
        log_scale=False,
        label="Signal region\npre-fit",
    )
    assert (
        compare_images("tests/visualize/reference/data_mc.png", str(fname), 0) is None
    )

    # log scale forced
    plot_model.data_mc(
        histo_dict_list_log,
        total_model_unc_log,
        bin_edges_log,
        figure_path=fname,
        log_scale=True,
        log_scale_x=True,
        label="Signal region\npre-fit",
    )
    assert (
        compare_images("tests/visualize/reference/data_mc_log.png", str(fname_log), 0)
        is None
    )

    # do not save figure, but close it
    # one bin with zero model prediction
    # custom histogram colors
    histo_dict_list[0]["yields"] = np.asarray([0, 14])
    histo_dict_list[1]["yields"] = np.asarray([0, 5])
    caplog.clear()
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        with pytest.warns() as warn_record:
            fig = plot_model.data_mc(
                histo_dict_list,
                total_model_unc_log,
                bin_edges_log,
                label="",
                colors={"Background": "blue", "Signal": "red"},
                close_figure=True,
            )
            assert mock_close_safe.call_args_list == [((fig, None, True), {})]
            # custom colors propagated to histogram
            assert (
                fig.axes[0].containers[0].patches[0].get_facecolor()
                == mpl.colors.to_rgba_array("blue")
            ).all()
            assert (
                fig.axes[0].containers[1].patches[0].get_facecolor()
                == mpl.colors.to_rgba_array("red")
            ).all()

    assert "predicted yield is zero in 1 bin(s), excluded from ratio plot" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # expect three RuntimeWarnings from numpy due to division by zero
    assert sum("divide by zero" in str(m.message) for m in warn_record) == 3

    # negative bin yield
    histo_dict_list[0]["yields"] = [-50, -50]
    with pytest.raises(
        ValueError,
        match=r"abc total model yield has negative bin\(s\): \[-50, -45\]",
    ):
        plot_model.data_mc(
            histo_dict_list, total_model_unc_log, bin_edges_log, label="abc"
        )

    plt.close("all")


def test_templates(tmp_path, caplog):
    fname = tmp_path / "fig.png"
    nominal_histo = {
        "yields": np.asarray([1.0, 1.2]),
        "stdev": np.asarray([0.05, 0.06]),
    }
    up_histo_orig = {
        "yields": np.asarray([1.2, 1.7]),
        "stdev": np.asarray([0.05, 0.07]),
    }
    down_histo_orig = {
        "yields": np.asarray([0.9, 0.9]),
        "stdev": np.asarray([0.06, 0.07]),
    }
    up_histo_mod = {"yields": np.asarray([1.3, 1.6]), "stdev": np.asarray([0.05, 0.07])}
    down_histo_mod = {
        "yields": np.asarray([0.85, 0.95]),
        "stdev": np.asarray([0.06, 0.07]),
    }
    bin_edges = np.asarray([0.0, 1.0, 2.0])
    variable = "x"
    label = "region: Signal region\nsample: Signal\nsystematic: Modeling"

    fig = plot_model.templates(
        nominal_histo,
        up_histo_orig,
        down_histo_orig,
        up_histo_mod,
        down_histo_mod,
        bin_edges,
        variable,
        figure_path=fname,
        label=label,
    )
    assert (
        compare_images("tests/visualize/reference/templates.png", str(fname), 0) is None
    )

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert (
        compare_images("tests/visualize/reference/templates.png", str(fname), 0) is None
    )
    caplog.clear()

    # do not save figure, but close it
    # only single variation specified
    # negative bin present in nominal histogram
    nominal_histo["yields"][0] = -1
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        fig = plot_model.templates(
            nominal_histo,
            up_histo_orig,
            {},
            up_histo_mod,
            {},
            bin_edges,
            variable,
            label=label,
            close_figure=True,
        )
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]
    assert (
        "region: Signal region\nsample: Signal\nsystematic: Modeling nominal histogram "
        "yield has negative bin(s): [-1.0, 1.2], taking absolute value for ratio plot "
        "uncertainty"
    ) in [rec.message for rec in caplog.records]
    caplog.clear()

    plt.close("all")


def test_modifier_grid(tmp_path):
    fname = tmp_path / "fig.png"
    grid_list = [
        np.asarray([[5, 4, 8, 2, 1, 3, 8], [7, 4, 8, 8, 8, 3, 0]]),
        np.asarray([[5, 8, 4, 2, 1, 3, 8], [6, 8, 4, 8, 8, 3, 0]]),
    ]
    axis_labels = [
        ["SR", "CR"],
        ["Background", "Signal"],
        ["a", "b", "c", "d", "e", "f", "mu"],
    ]
    category_map = {
        0: "normfactor",
        1: "shapefactor",
        2: "shapesys",
        3: "lumi",
        4: "staterror",
        5: "normsys + histosys",
        6: "histosys",
        7: "normsys",
        8: "none",
    }

    fig = plot_model.modifier_grid(
        grid_list, axis_labels, category_map, figure_path=fname
    )
    # non-zero tolerance as layout changes very slightly in CI
    assert (
        compare_images("tests/visualize/reference/modifier_grid.png", str(fname), 16)
        is None
    )

    # saving the return figure again seems to slightly change formatting, so instead of
    # doing that here in the test (like in the tests above), do not save the figure
    # automatically and only manually save it in the next test to compare

    # do not save figure, but close it and save returned figure manually
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        fig = plot_model.modifier_grid(
            grid_list, axis_labels, category_map, close_figure=True
        )
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]

    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert (
        compare_images("tests/visualize/reference/modifier_grid.png", str(fname), 16)
        is None
    )

    plt.close("all")



import logging
import re
from unittest import mock

import iminuit
import numpy as np
import pyhf
import pytest

from cabinetry import fit
from cabinetry import model_utils


def test_print_results(caplog):
    caplog.set_level(logging.DEBUG)

    bestfit = np.asarray([1.0, 2.0])
    uncertainty = np.asarray([0.1, 0.3])
    labels = ["param_A", "param_B"]
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 0.0)

    fit.print_results(fit_results)
    assert "param_A =  1.0000 +/- 0.1000" in [rec.message for rec in caplog.records]
    assert "param_B =  2.0000 +/- 0.3000" in [rec.message for rec in caplog.records]
    caplog.clear()

    # constant parameter
    uncertainty = np.asarray([0.1, 0.0])
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 0.0)
    fit.print_results(fit_results)
    assert "param_B =  2.0000 +/- 0.0000  (constant)" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()


# skip a "RuntimeWarning: numpy.ufunc size changed" warning
# due to different numpy versions used in dependencies
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@mock.patch("cabinetry.fit._run_minos", return_value={"Signal strength": (-0.1, 0.2)})
def test__fit_model_pyhf(mock_minos, example_spec, example_spec_multibin):
    pyhf.set_backend("numpy", "scipy")
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_pyhf(model, data)
    assert np.allclose(fit_results.bestfit, [8.33624084, 1.1])
    assert np.allclose(fit_results.uncertainty, [0.38182003, 0.0])
    assert fit_results.labels == ["Signal strength", "staterror_Signal-Region[0]"]
    assert np.allclose(fit_results.best_twice_nll, 7.82495235)
    assert np.allclose(fit_results.corr_mat, [[1.0, 0.0], [0.0, 0.0]])
    assert pyhf.get_backend()[1].name == "scipy"  # optimizer was reset

    # Asimov fit, with fixed gamma (fixed not to Asimov MLE)
    model, data = model_utils.model_and_data(example_spec, asimov=True)
    fit_results = fit._fit_model_pyhf(model, data)
    # the gamma factor is multiplicative and fixed to 1.1, so the
    # signal strength needs to be 1/1.1 to compensate
    assert np.allclose(fit_results.bestfit, [0.90917877, 1.1])
    assert np.allclose(fit_results.uncertainty, [0.12628017, 0.0])
    assert fit_results.labels == ["Signal strength", "staterror_Signal-Region[0]"]
    assert np.allclose(fit_results.best_twice_nll, 5.61189476)
    assert np.allclose(fit_results.corr_mat, [[1.0, 0.0], [0.0, 0.0]])

    # parameters held constant via keyword argument
    model, data = model_utils.model_and_data(example_spec_multibin)
    init_pars = model.config.suggested_init()
    init_pars[1] = 0.9
    init_pars[2] = 1.1
    fix_pars = model.config.suggested_fixed()
    fix_pars[1] = True
    fix_pars[2] = True
    fit_results = fit._fit_model_pyhf(
        model, data, init_pars=init_pars, fix_pars=fix_pars
    )
    assert np.allclose(fit_results.bestfit, [1.48041923, 0.9, 1.1, 0.97511112])
    assert np.allclose(fit_results.uncertainty, [0.20694409, 0.0, 0.0, 0.11792805])
    assert np.allclose(fit_results.best_twice_nll, 10.4531891)

    # custom parameter bounds
    model, data = model_utils.model_and_data(example_spec)
    par_bounds = [(0, 5), (0.1, 2)]
    fit_results = fit._fit_model_pyhf(model, data, par_bounds=par_bounds)
    assert np.allclose(fit_results.bestfit, [5.0, 1.1])

    # propagation of strategy, max iterations, tolerance
    model, data = model_utils.model_and_data(example_spec)
    with mock.patch("pyhf.infer.mle.fit") as mock_fit:
        # mock return value will cause ValueError immediately after call
        # could alternatively use mocker.spy from pytest-mock
        with pytest.raises(ValueError):
            fit._fit_model_pyhf(model, data)
        # strategy kwarg not used in call, so not propagated to pyhf
        assert "strategy" not in mock_fit.call_args[1].keys()
        assert mock_fit.call_args[1]["maxiter"] is None
        assert mock_fit.call_args[1]["tolerance"] is None

        with pytest.raises(ValueError):
            fit._fit_model_pyhf(model, data, strategy=2, maxiter=100, tolerance=0.01)
        assert mock_fit.call_args[1]["strategy"] == 2
        assert mock_fit.call_args[1]["maxiter"] == 100
        assert mock_fit.call_args[1]["tolerance"] == 0.01

    # including minos, one parameter is unknown
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_pyhf(model, data, minos=["Signal strength", "abc"])
    assert fit_results.minos_uncertainty["Signal strength"] == (-0.1, 0.2)
    assert mock_minos.call_count == 1
    # first argument to minos call is the Minuit instance
    assert mock_minos.call_args[0][1] == ["Signal strength", "abc"]
    assert mock_minos.call_args[0][2] == [
        "Signal strength",
        "staterror_Signal-Region[0]",
    ]
    assert mock_minos.call_args[1] == {}


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@mock.patch("cabinetry.fit._run_minos", return_value={"Signal strength": (-0.1, 0.2)})
def test__fit_model_custom(mock_minos, example_spec, example_spec_multibin):
    pyhf.set_backend("numpy", "scipy")
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_custom(model, data)
    assert np.allclose(fit_results.bestfit, [8.33625071, 1.1])
    assert np.allclose(fit_results.uncertainty, [0.38182151, 0.0])
    assert fit_results.labels == ["Signal strength", "staterror_Signal-Region[0]"]
    assert np.allclose(fit_results.best_twice_nll, 7.82495235)
    assert np.allclose(fit_results.corr_mat, [[1.0, 0.0], [0.0, 0.0]])
    assert pyhf.get_backend()[1].name == "scipy"  # optimizer was reset

    # Asimov fit, with fixed gamma (fixed not to Asimov MLE)
    model, data = model_utils.model_and_data(example_spec, asimov=True)
    fit_results = fit._fit_model_custom(model, data)
    # the gamma factor is multiplicative and fixed to 1.1, so the
    # signal strength needs to be 1/1.1 to compensate
    assert np.allclose(fit_results.bestfit, [0.90917877, 1.1])
    assert np.allclose(fit_results.uncertainty, [0.12628023, 0.0])
    assert fit_results.labels == ["Signal strength", "staterror_Signal-Region[0]"]
    assert np.allclose(fit_results.best_twice_nll, 5.61189476)
    assert np.allclose(fit_results.corr_mat, [[1.0, 0.0], [0.0, 0.0]])

    # parameters held constant via keyword argument
    model, data = model_utils.model_and_data(example_spec_multibin)
    init_pars = model.config.suggested_init()
    init_pars[1] = 0.9
    init_pars[2] = 1.1
    fix_pars = model.config.suggested_fixed()
    fix_pars[1] = True
    fix_pars[2] = True
    fit_results = fit._fit_model_custom(
        model, data, init_pars=init_pars, fix_pars=fix_pars
    )
    assert np.allclose(fit_results.bestfit, [1.48041923, 0.9, 1.1, 0.97511112])
    assert np.allclose(fit_results.uncertainty, [0.20694409, 0.0, 0.0, 0.11792805])
    assert np.allclose(fit_results.best_twice_nll, 10.45318909)

    # custom parameter bounds
    model, data = model_utils.model_and_data(example_spec)
    par_bounds = [(0, 5), (0.1, 2)]
    fit_results = fit._fit_model_custom(model, data, par_bounds=par_bounds)
    assert np.allclose(fit_results.bestfit, [5.0, 1.1])

    # propagation of strategy, max iterations, tolerance
    model, data = model_utils.model_and_data(example_spec)
    mock_minuit_instance = mock.MagicMock()
    mock_minuit_instance.errors = [0.1, 0.1]  # needed as it will be accessed
    with mock.patch("iminuit.Minuit", return_value=mock_minuit_instance):
        # mocked minuit instance used to check correct propagation of settings
        fit._fit_model_custom(model, data)
        assert mock_minuit_instance.strategy == 1  # default for numpy backend
        assert mock_minuit_instance.migrad.call_args == ((), {"ncall": 100_000})
        assert mock_minuit_instance.tol == 0.1

        fit._fit_model_custom(model, data, strategy=2, maxiter=100, tolerance=0.01)
        assert mock_minuit_instance.strategy == 2
        assert mock_minuit_instance.migrad.call_args == ((), {"ncall": 100})
        assert mock_minuit_instance.tol == 0.01

    # failed fit (simulated via limited number of iterations)
    with pytest.raises(ValueError, match="Minimization failed, minimum is invalid."):
        fit._fit_model_custom(model, data, maxiter=10)

    # including minos
    model, data = model_utils.model_and_data(example_spec)
    fit_results = fit._fit_model_custom(model, data, minos=["Signal strength"])
    assert fit_results.minos_uncertainty["Signal strength"] == (-0.1, 0.2)
    assert mock_minos.call_count == 1
    # first argument to minos call is the Minuit instance
    assert mock_minos.call_args[0][1] == ["Signal strength"]
    assert mock_minos.call_args[0][2] == [
        "Signal strength",
        "staterror_Signal-Region[0]",
    ]
    assert mock_minos.call_args[1] == {}


@mock.patch(
    "cabinetry.fit._fit_model_custom",
    return_value=fit.FitResults(
        np.asarray([1.2]), np.asarray([0.2]), ["par"], np.empty(0), 2.0
    ),
)
@mock.patch(
    "cabinetry.fit._fit_model_pyhf",
    return_value=fit.FitResults(
        np.asarray([1.1]), np.asarray([0.2]), ["par"], np.empty(0), 2.0
    ),
)
def test__fit_model(mock_pyhf, mock_custom, example_spec):
    model, data = model_utils.model_and_data(example_spec)

    # pyhf API
    fit_results = fit._fit_model(model, data)
    assert mock_pyhf.call_count == 1
    assert mock_pyhf.call_args[0][0].spec == model.spec
    assert mock_pyhf.call_args[0][1] == data
    assert mock_pyhf.call_args[1] == {
        "minos": None,
        "init_pars": None,
        "fix_pars": None,
        "par_bounds": None,
        "strategy": None,
        "maxiter": None,
        "tolerance": None,
    }
    assert np.allclose(fit_results.bestfit, [1.1])

    # pyhf API, init/fixed pars, par bounds, minos, strategy/maxiter/tolerance
    fit_results = fit._fit_model(
        model,
        data,
        minos=["Signal strength"],
        init_pars=[1.5, 2.0],
        fix_pars=[False, True],
        par_bounds=[(0, 5), (0.1, 10.0)],
        strategy=2,
        maxiter=100,
        tolerance=0.01,
    )
    assert mock_pyhf.call_count == 2
    assert mock_pyhf.call_args[0][0].spec == model.spec
    assert mock_pyhf.call_args[0][1] == data
    assert mock_pyhf.call_args[1] == {
        "minos": ["Signal strength"],
        "init_pars": [1.5, 2.0],
        "fix_pars": [False, True],
        "par_bounds": [(0, 5), (0.1, 10.0)],
        "strategy": 2,
        "maxiter": 100,
        "tolerance": 0.01,
    }
    assert np.allclose(fit_results.bestfit, [1.1])

    # direct iminuit
    fit_results = fit._fit_model(model, data, custom_fit=True)
    assert mock_custom.call_count == 1
    assert mock_custom.call_args[0][0].spec == model.spec
    assert mock_custom.call_args[0][1] == data
    assert mock_custom.call_args[1] == {
        "minos": None,
        "init_pars": None,
        "fix_pars": None,
        "par_bounds": None,
        "strategy": None,
        "maxiter": None,
        "tolerance": None,
    }
    assert np.allclose(fit_results.bestfit, [1.2])

    # direct iminuit, init/fixed pars, par bounds, minos, strategy/maxiter/tolerance
    fit_results = fit._fit_model(
        model,
        data,
        minos=["Signal strength"],
        init_pars=[1.5, 2.0],
        fix_pars=[False, True],
        par_bounds=[(0, 5), (0.1, 10.0)],
        strategy=2,
        maxiter=100,
        tolerance=0.01,
        custom_fit=True,
    )
    assert mock_custom.call_count == 2
    assert mock_custom.call_args[0][0].spec == model.spec
    assert mock_custom.call_args[0][1] == data
    assert mock_custom.call_args[1] == {
        "minos": ["Signal strength"],
        "init_pars": [1.5, 2.0],
        "fix_pars": [False, True],
        "par_bounds": [(0, 5), (0.1, 10.0)],
        "strategy": 2,
        "maxiter": 100,
        "tolerance": 0.01,
    }
    assert np.allclose(fit_results.bestfit, [1.2])


def test__run_minos(caplog):
    caplog.set_level(logging.DEBUG)

    def func_to_minimize(pars):
        # mock NLL
        return (
            np.sum(
                np.power(pars - 2 * np.ones_like(pars), 2)
                + np.power(pars - 1 * np.ones_like(pars), 4)
            )
            + pars[0]
        )

    m = iminuit.Minuit(func_to_minimize, [1.0, 1.0], name=["a", "b"])
    m.errordef = 1
    m.migrad()

    minos_results = fit._run_minos(m, ["b"], ["a", "b"])
    assert np.allclose(minos_results["b"][0], -0.72622053)
    assert np.allclose(minos_results["b"][1], 0.47381869)
    assert "running MINOS for b" in [rec.message for rec in caplog.records]
    assert "b =  1.5909 -0.7262 +0.4738" in [rec.message for rec in caplog.records]
    caplog.clear()

    # unknown parameter, MINOS does not run
    m = iminuit.Minuit(func_to_minimize, [1.0, 1.0])
    m.errordef = 1
    m.migrad()

    minos_results = fit._run_minos(m, ["x2"], ["a", "b"])
    assert minos_results == {}
    assert [rec.message for rec in caplog.records] == [
        "parameter x2 not found in model",
        "MINOS results:",
    ]
    caplog.clear()


@mock.patch("cabinetry.model_utils.unconstrained_parameter_count", return_value=1)
@mock.patch(
    "cabinetry.model_utils._parameters_maximizing_constraint_term",
    side_effect=[
        [1.0, 1.0, 1.0, 1.0],  # for example_spec_multibin
        [1.0, 0.9, 1.1, 0.8],  # for custom aux
        [1.0],  # for example_spec_no_aux
    ],
)
def test__goodness_of_fit(
    mock_pars, mock_count, example_spec_multibin, example_spec_no_aux, caplog
):
    caplog.set_level(logging.DEBUG)

    model, data = model_utils.model_and_data(example_spec_multibin)
    p_val = fit._goodness_of_fit(model, data, 9.964913)
    assert mock_pars.call_count == 1
    assert mock_pars.call_args[0][0].spec == model.spec
    assert np.allclose(mock_pars.call_args[0][1], [1.0, 1.0, 1.0])
    assert mock_pars.call_args[1] == {}
    assert mock_count.call_count == 1
    assert mock_count.call_args[0][0].spec == model.spec
    assert mock_count.call_args[1] == {}
    assert "Delta NLL = 0.084185" in [rec.message for rec in caplog.records]
    assert np.allclose(p_val, 0.91926079)
    caplog.clear()

    # same setup but using custom auxdata
    model, _ = model_utils.model_and_data(example_spec_multibin)
    data = [35, 8, 10] + [0.9, 1.1, 0.8]  # custom aux
    p_val = fit._goodness_of_fit(model, data, 9.964913)
    assert mock_pars.call_count == 2
    assert np.allclose(mock_pars.call_args[0][1], [0.9, 1.1, 0.8])  # aux picked up
    assert np.allclose(p_val, 0.91926079)  # same result as before

    # no auxdata and zero degrees of freedom in chi2 test
    model, data = model_utils.model_and_data(example_spec_no_aux)
    p_val = fit._goodness_of_fit(model, data, 6.01482863)
    assert mock_pars.call_count == 2  # no new call (no auxdata)
    assert mock_count.call_count == 3
    assert mock_count.call_args[0][0].spec == model.spec
    assert mock_count.call_args[1] == {}
    assert (
        "cannot calculate p-value: 0 degrees of freedom and Delta NLL = 0.000000"
        in [rec.message for rec in caplog.records]
    )
    assert np.isnan(p_val)
    caplog.clear()


@mock.patch("cabinetry.fit._goodness_of_fit", return_value=0.1)
@mock.patch("cabinetry.fit.print_results")
@mock.patch(
    "cabinetry.fit._fit_model",
    return_value=fit.FitResults(
        np.asarray([1.0]), np.asarray([0.1]), ["par"], np.empty(0), 2.0
    ),
)
def test_fit(mock_fit, mock_print, mock_gof):
    model = mock.MagicMock()
    data = mock.MagicMock()

    # fit through pyhf.infer API
    fit_results = fit.fit(model, data)
    assert mock_fit.call_args_list == [
        (
            (model, data),
            {
                "minos": None,
                "init_pars": None,
                "fix_pars": None,
                "par_bounds": None,
                "strategy": None,
                "maxiter": None,
                "tolerance": None,
                "custom_fit": False,
            },
        )
    ]
    mock_print.assert_called_once()
    assert mock_print.call_args[0][0].bestfit == [1.0]
    assert mock_print.call_args[0][0].uncertainty == [0.1]
    assert mock_print.call_args[0][0].labels == ["par"]
    assert mock_gof.call_count == 0
    assert fit_results.bestfit == [1.0]

    # custom fit, init/fix pars, par bounds, strategy/maxiter/tolerance
    init_pars = [2.0]
    fix_pars = [True]
    par_bounds = [(0.0, 5.0)]
    fit_results = fit.fit(
        model,
        data,
        init_pars=init_pars,
        fix_pars=fix_pars,
        par_bounds=par_bounds,
        strategy=2,
        maxiter=100,
        tolerance=0.01,
        custom_fit=True,
    )
    assert mock_fit.call_count == 2
    assert mock_fit.call_args == (
        (model, data),
        {
            "minos": None,
            "init_pars": init_pars,
            "fix_pars": fix_pars,
            "par_bounds": par_bounds,
            "strategy": 2,
            "maxiter": 100,
            "tolerance": 0.01,
            "custom_fit": True,
        },
    )
    assert mock_print.call_args[0][0].bestfit == [1.0]
    assert mock_print.call_args[0][0].uncertainty == [0.1]
    assert fit_results.bestfit == [1.0]

    # parameters for MINOS
    fit_results = fit.fit(model, data, minos=["abc"])
    assert mock_fit.call_count == 3
    assert mock_fit.call_args[1] == {
        "minos": ["abc"],
        "init_pars": None,
        "fix_pars": None,
        "par_bounds": None,
        "strategy": None,
        "maxiter": None,
        "tolerance": None,
        "custom_fit": False,
    }
    assert fit_results.bestfit == [1.0]
    fit_results = fit.fit(model, data, minos="abc", custom_fit=True)
    assert mock_fit.call_count == 4
    assert mock_fit.call_args[1] == {
        "minos": ["abc"],
        "init_pars": None,
        "fix_pars": None,
        "par_bounds": None,
        "strategy": None,
        "maxiter": None,
        "tolerance": None,
        "custom_fit": True,
    }
    assert fit_results.bestfit == [1.0]

    # goodness-of-fit test
    fit_results_gof = fit.fit(model, data, goodness_of_fit=True)
    assert mock_gof.call_args_list == [((model, data, 2.0), {})]
    assert fit_results_gof.goodness_of_fit == 0.1


@mock.patch(
    "cabinetry.fit._fit_model",
    side_effect=[
        fit.FitResults(
            np.asarray([1.3, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.7, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([1.2, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.8, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        # for second ranking call with fixed parameter
        fit.FitResults(
            np.asarray([1.2, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.8, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        # for third ranking call without reference results
        fit.FitResults(
            np.asarray([1.0, 0.9]), np.asarray([0.3, 0.3]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([1.3, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
        fit.FitResults(
            np.asarray([0.7, 0.9]), np.asarray([0.1, 0.1]), ["a", "b"], np.empty(0), 0.0
        ),
    ],
)
def test_ranking(mock_fit, example_spec):
    example_spec["measurements"][0]["config"]["parameters"][0]["fixed"] = False
    bestfit = np.asarray([1.0, 0.9])
    uncertainty = np.asarray([0.1, 0.02])
    labels = ["Signal strength", "staterror"]
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 0.0)
    model, data = model_utils.model_and_data(example_spec)
    ranking_results = fit.ranking(model, data, fit_results=fit_results)

    # correct call to fit
    expected_fix = [False, True]
    expected_inits = [[1.0, 0.95019305], [1.0, 0.84980695], [1.0, 0.92], [1.0, 0.88]]
    assert mock_fit.call_count == 4
    for i in range(4):
        assert mock_fit.call_args_list[i][0] == (model, data)
        assert np.allclose(
            mock_fit.call_args_list[i][1]["init_pars"], expected_inits[i]
        )
        assert np.allclose(mock_fit.call_args_list[i][1]["fix_pars"], expected_fix)
        assert mock_fit.call_args_list[i][1]["par_bounds"] is None
        assert mock_fit.call_args_list[i][1]["strategy"] is None
        assert mock_fit.call_args_list[i][1]["maxiter"] is None
        assert mock_fit.call_args_list[i][1]["tolerance"] is None
        assert mock_fit.call_args_list[i][1]["custom_fit"] is False

    # POI removed from fit results
    assert np.allclose(ranking_results.bestfit, [0.9])
    assert np.allclose(ranking_results.uncertainty, [0.02])
    assert ranking_results.labels == ["staterror"]

    # received correct mock results
    assert np.allclose(ranking_results.prefit_up, [0.3])
    assert np.allclose(ranking_results.prefit_down, [-0.3])
    assert np.allclose(ranking_results.postfit_up, [0.2])
    assert np.allclose(ranking_results.postfit_down, [-0.2])

    # fixed parameter in ranking, custom fit, POI via kwarg
    example_spec["measurements"][0]["config"]["parameters"][0]["fixed"] = True
    example_spec["measurements"][0]["config"]["poi"] = ""
    model, data = model_utils.model_and_data(example_spec)
    ranking_results = fit.ranking(
        model,
        data,
        fit_results=fit_results,
        poi_name="Signal strength",
        custom_fit=True,
    )
    # expect two calls in this ranking (and had 4 before, so 6 total): pre-fit
    # uncertainty is 0 since parameter is fixed, mock post-fit uncertainty is not 0
    assert mock_fit.call_count == 6
    assert mock_fit.call_args[1]["custom_fit"] is True
    assert np.allclose(ranking_results.prefit_up, [0.0])
    assert np.allclose(ranking_results.prefit_down, [0.0])
    assert np.allclose(ranking_results.postfit_up, [0.2])
    assert np.allclose(ranking_results.postfit_down, [-0.2])

    # no reference results, init/fixed pars, par bounds, strategy/maxiter/tolerance
    ranking_results = fit.ranking(
        model,
        data,
        poi_name="Signal strength",
        init_pars=[1.5, 1.0],
        fix_pars=[False, False],
        par_bounds=[(0, 5), (0.1, 10)],
        strategy=2,
        maxiter=100,
        tolerance=0.01,
        custom_fit=True,
    )
    assert mock_fit.call_count == 9
    # reference fit
    assert mock_fit.call_args_list[-3] == (
        (model, data),
        {
            "init_pars": [1.5, 1.0],
            "fix_pars": [False, False],
            "par_bounds": [(0, 5), (0.1, 10)],
            "strategy": 2,
            "maxiter": 100,
            "tolerance": 0.01,
            "custom_fit": True,
        },
    )
    # fits for impact (comparing each option separately since init_pars needs allclose)
    assert mock_fit.call_args_list[-2][0] == (model, data)
    assert np.allclose(mock_fit.call_args_list[-2][1]["init_pars"], [1.5, 1.2])
    assert mock_fit.call_args_list[-2][1]["fix_pars"] == [False, True]
    assert mock_fit.call_args_list[-2][1]["par_bounds"] == [(0, 5), (0.1, 10)]
    assert mock_fit.call_args_list[-2][1]["strategy"] == 2
    assert mock_fit.call_args_list[-2][1]["maxiter"] == 100
    assert mock_fit.call_args_list[-2][1]["tolerance"] == 0.01
    assert mock_fit.call_args_list[-2][1]["custom_fit"] is True
    assert mock_fit.call_args_list[-1][0] == (model, data)
    assert np.allclose(mock_fit.call_args_list[-1][1]["init_pars"], [1.5, 0.6])
    assert mock_fit.call_args_list[-1][1]["fix_pars"] == [False, True]
    assert mock_fit.call_args_list[-1][1]["par_bounds"] == [(0, 5), (0.1, 10)]
    assert mock_fit.call_args_list[-1][1]["strategy"] == 2
    assert mock_fit.call_args_list[-1][1]["maxiter"] == 100
    assert mock_fit.call_args_list[-1][1]["tolerance"] == 0.01
    assert mock_fit.call_args_list[-1][1]["custom_fit"] is True
    # ranking results
    assert np.allclose(ranking_results.prefit_up, [0.0])
    assert np.allclose(ranking_results.prefit_down, [0.0])
    assert np.allclose(ranking_results.postfit_up, [0.3])
    assert np.allclose(ranking_results.postfit_down, [-0.3])

    # no POI specified anywhere
    with pytest.raises(ValueError, match="no POI specified, cannot calculate ranking"):
        fit.ranking(model, data, fit_results=fit_results)


@mock.patch(
    "cabinetry.fit._fit_model",
    side_effect=[
        fit.FitResults(
            np.asarray([1.3, 0.9]), np.asarray([0.1, 0.1]), [], np.empty(0), 8.0
        )
    ]  # nominal fit
    + [
        fit.FitResults(np.empty(0), np.empty(0), [], np.empty(0), abs(i) + 8)
        for i in np.linspace(-5, 5, 11)
    ]  # fits in scan
    + [
        fit.FitResults(
            np.asarray([1.3, 0.9]), np.asarray([0.1, 0.1]), [], np.empty(0), 2.0
        )
    ]
    * 6,  # fits for custom parameter range
)
def test_scan(mock_fit, example_spec):
    expected_scan_values = np.linspace(1.1, 1.5, 11)
    # -2 log(L) from unconstrained fit subtracted from expected NLLs
    expected_delta_nlls = np.abs(np.linspace(-5, 5, 11))
    model, data = model_utils.model_and_data(example_spec)

    par_name = "Signal strength"
    scan_results = fit.scan(model, data, par_name)
    assert scan_results.name == par_name
    assert scan_results.bestfit == 1.3
    assert scan_results.uncertainty == 0.1
    assert np.allclose(scan_results.parameter_values, expected_scan_values)
    assert np.allclose(scan_results.delta_nlls, expected_delta_nlls)

    assert mock_fit.call_count == 12
    # unconstrained fit
    assert mock_fit.call_args_list[0][0] == ((model, data))
    assert mock_fit.call_args_list[0][1] == {
        "init_pars": None,
        "fix_pars": None,
        "par_bounds": None,
        "strategy": None,
        "maxiter": None,
        "tolerance": None,
        "custom_fit": False,
    }
    # fits in scan
    for i, scan_val in enumerate(expected_scan_values):
        assert mock_fit.call_args_list[i + 1][0] == ((model, data))
        assert mock_fit.call_args_list[i + 1][1] == {
            "init_pars": [scan_val, 1.1],
            "fix_pars": [True, True],
            "par_bounds": None,
            "strategy": None,
            "maxiter": None,
            "tolerance": None,
            "custom_fit": False,
        }

    # parameter range specified, custom fit, init/fixed pars, par bounds,
    # strategy/maxiter/tolerance
    scan_results = fit.scan(
        model,
        data,
        par_name,
        par_range=(1.0, 1.5),
        n_steps=5,
        init_pars=[1.0, 1.0],
        fix_pars=[False, False],
        par_bounds=[(0, 5), (0.1, 10)],
        strategy=2,
        maxiter=100,
        tolerance=0.01,
        custom_fit=True,
    )
    expected_custom_scan = np.linspace(1.0, 1.5, 5)
    assert np.allclose(scan_results.parameter_values, expected_custom_scan)
    assert mock_fit.call_args[1]["custom_fit"] is True
    assert mock_fit.call_args[1] == {
        "init_pars": [1.5, 1.0],  # last step of scan
        "fix_pars": [True, False],
        "par_bounds": [(0, 5), (0.1, 10)],
        "strategy": 2,
        "maxiter": 100,
        "tolerance": 0.01,
        "custom_fit": True,
    }

    # unknown parameter
    with pytest.raises(ValueError, match="parameter abc not found in model"):
        fit.scan(model, data, "abc")


def test_limit(example_spec_with_background, caplog):
    caplog.set_level(logging.DEBUG)
    pyhf.set_backend("numpy", "scipy")

    # expected values for results
    observed_limit = 0.749
    expected_limit = [0.303, 0.411, 0.581, 0.833, 1.160]

    # modify workspace to include custom POI range
    example_spec_with_background["measurements"][0]["config"]["parameters"][0][
        "bounds"
    ] = [[0, 8]]
    model, data = model_utils.model_and_data(example_spec_with_background)

    limit_results = fit.limit(model, data)
    assert np.allclose(limit_results.observed_limit, observed_limit, rtol=1e-2)
    assert np.allclose(limit_results.expected_limit, expected_limit, rtol=1e-2)
    # compare a few CLs values
    assert np.allclose(limit_results.observed_CLs[0], 0.780874)
    assert np.allclose(
        limit_results.expected_CLs[0],
        [0.402421, 0.548538, 0.719383, 0.878530, 0.971678],
    )
    assert np.allclose(limit_results.poi_values[0], 0.1)
    assert np.allclose(limit_results.observed_CLs[-1], 0.0)
    assert np.allclose(limit_results.expected_CLs[-1], [0.0, 0.0, 0.0, 0.0, 0.0])
    assert np.allclose(limit_results.poi_values[-1], 8.0)  # from custom POI range
    assert limit_results.confidence_level == 0.95
    # verify that POI values are sorted
    assert np.allclose(limit_results.poi_values, sorted(limit_results.poi_values))
    assert pyhf.get_backend()[1].name == "scipy"  # optimizer was reset
    caplog.clear()

    # access negative POI values with lower bracket below zero
    limit_results = fit.limit(model, data, bracket=(-1, 5), poi_tolerance=0.05)
    assert "skipping fit for Signal strength = -1.0000, setting CLs = 1" in [
        rec.message for rec in caplog.records
    ]
    assert np.allclose(limit_results.observed_limit, observed_limit, rtol=5e-2)
    assert np.allclose(limit_results.expected_limit, expected_limit, rtol=5e-2)
    caplog.clear()

    # convergence issues due to number of iterations
    fit.limit(model, data, bracket=(0.1, 1), maxsteps=1)
    assert "one or more calculations did not converge, check log" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # Asimov dataset with nominal signal strength of 0, POI via kwarg
    example_spec_with_background["measurements"][0]["config"]["parameters"][0][
        "inits"
    ] = [0.0]
    example_spec_with_background["measurements"][0]["config"]["poi"] = ""
    model, data = model_utils.model_and_data(example_spec_with_background, asimov=True)
    assert model.config.poi_index is None  # no POI set before calculation
    assert model.config.poi_name is None
    limit_results = fit.limit(model, data, poi_name="Signal strength")
    assert np.allclose(limit_results.observed_limit, 0.586, rtol=2e-2)
    assert np.allclose(limit_results.expected_limit, expected_limit, rtol=2e-2)
    assert model.config.poi_index is None  # model config is preserved
    assert model.config.poi_name is None
    caplog.clear()

    # bracket does not contain root, custom confidence level
    with pytest.raises(
        ValueError, match=re.escape("f(a) and f(b) must have different signs")
    ):
        fit.limit(
            model,
            data,
            bracket=(1.0, 2.0),
            confidence_level=0.9,
            poi_name="Signal strength",
        )
    assert (
        "CLs values at 1.0000 and 2.0000 do not bracket CLs=0.1000, try a different "
        "starting bracket" in [rec.message for rec in caplog.records]
    )
    caplog.clear()

    # bracket with identical values
    with pytest.raises(ValueError, match="the two bracket values must not be the same"):
        fit.limit(model, data, bracket=(3.0, 3.0), poi_name="Signal strength")
    caplog.clear()

    # init/fixed pars, par bounds, lower POI bound below 0
    with mock.patch("pyhf.infer.hypotest", return_value=None) as mock_test:
        # mock return value will cause TypeError immediately after call
        # could alternatively use mocker.spy from pytest-mock
        with pytest.raises(TypeError):
            fit.limit(
                model,
                data,
                poi_name="Signal strength",
                init_pars=[0.9, 1.0],
                fix_pars=[False, True],
                par_bounds=[(-1, 5), (0.1, 10.0)],
            )
        assert mock_test.call_args_list == [
            (
                (0.1, data, model),
                {
                    "init_pars": [0.9, 1.0],
                    "fixed_params": [False, True],
                    "par_bounds": [(0, 5), (0.1, 10.0)],
                    "test_stat": "qtilde",
                    "return_expected_set": True,
                },
            )
        ]
    assert "setting lower parameter bound for POI to 0" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # new model, error raised in previous step caused changes in poi_index / poi_name
    model, data = model_utils.model_and_data(example_spec_with_background)

    # no POI specified anywhere
    with pytest.raises(ValueError, match="no POI specified, cannot calculate limit"):
        fit.limit(model, data)

    # add POI back to model and reset backend for testing optimizer customization
    example_spec_with_background["measurements"][0]["config"]["poi"] = "Signal strength"
    model, data = model_utils.model_and_data(example_spec_with_background)
    pyhf.set_backend("numpy", "scipy")

    # default strategy/maxiter/tolerance
    with mock.patch("pyhf.set_backend") as mock_backend:
        fit.limit(model, data)
    assert mock_backend.call_count == 2
    # setting to minuit
    assert mock_backend.call_args_list[0][0][1].name == "minuit"
    assert mock_backend.call_args_list[0][0][1].verbose == 1
    assert mock_backend.call_args_list[0][0][1].strategy is None
    assert mock_backend.call_args_list[0][0][1].maxiter is None
    assert mock_backend.call_args_list[0][0][1].tolerance is None
    assert mock_backend.call_args_list[0][1] == {}
    # resetting back to scipy
    assert mock_backend.call_args_list[1][0][1].name == "scipy"

    # custom strategy/maxiter/tolerance
    with mock.patch("pyhf.set_backend") as mock_backend:
        fit.limit(model, data, strategy=2, maxiter=100, tolerance=0.01)
    assert mock_backend.call_count == 2
    # setting to minuit
    assert mock_backend.call_args_list[0][0][1].name == "minuit"
    assert mock_backend.call_args_list[0][0][1].verbose == 1
    assert mock_backend.call_args_list[0][0][1].strategy == 2
    assert mock_backend.call_args_list[0][0][1].maxiter == 100
    assert mock_backend.call_args_list[0][0][1].tolerance == 0.01
    assert mock_backend.call_args_list[0][1] == {}
    # resetting back to scipy
    assert mock_backend.call_args_list[1][0][1].name == "scipy"


def test_significance(example_spec_with_background):
    pyhf.set_backend("numpy", "scipy")
    # increase observed data for smaller observed p-value
    example_spec_with_background["observations"][0]["data"] = [196]

    model, data = model_utils.model_and_data(example_spec_with_background)
    significance_results = fit.significance(model, data)
    assert np.allclose(significance_results.observed_p_value, 0.00080517)
    assert np.allclose(significance_results.observed_significance, 3.15402672)
    assert np.allclose(significance_results.expected_p_value, 0.00033333)
    assert np.allclose(significance_results.expected_significance, 3.40293444)
    assert pyhf.get_backend()[1].name == "scipy"  # optimizer was reset

    # reduce signal for larger expected p-value
    example_spec_with_background["channels"][0]["samples"][0]["data"] = [30]

    # Asimov dataset, observed = expected, POI removed from measurement config
    example_spec_with_background["measurements"][0]["config"]["poi"] = ""
    model, data = model_utils.model_and_data(example_spec_with_background, asimov=True)
    assert model.config.poi_index is None  # no POI set before calculation
    assert model.config.poi_name is None
    significance_results = fit.significance(model, data, poi_name="Signal strength")
    assert np.allclose(significance_results.observed_p_value, 0.02062714)
    assert np.allclose(significance_results.observed_significance, 2.04096523)
    assert np.allclose(significance_results.expected_p_value, 0.02062714)
    assert np.allclose(significance_results.expected_significance, 2.04096523)
    assert model.config.poi_index is None  # model config is preserved
    assert model.config.poi_name is None

    # init/fixed pars, par bounds
    model, data = model_utils.model_and_data(example_spec_with_background)
    with mock.patch("pyhf.infer.hypotest", return_value=(0.0, 0.0)) as mock_test:
        fit.significance(
            model,
            data,
            poi_name="Signal strength",
            init_pars=[0.9, 1.0],
            fix_pars=[False, True],
            par_bounds=[(0, 5), (0.1, 10.0)],
        )
    assert mock_test.call_args_list == [
        (
            (0.0, data, model),
            {
                "init_pars": [0.9, 1.0],
                "fixed_params": [False, True],
                "par_bounds": [(0, 5), (0.1, 10.0)],
                "test_stat": "q0",
                "return_expected": True,
            },
        )
    ]

    # no POI specified anywhere
    with pytest.raises(
        ValueError, match="no POI specified, cannot calculate significance"
    ):
        fit.significance(model, data)

    # add POI back to model and reset backend for testing optimizer customization
    example_spec_with_background["measurements"][0]["config"]["poi"] = "Signal strength"
    model, data = model_utils.model_and_data(example_spec_with_background)
    pyhf.set_backend("numpy", "scipy")

    # default strategy/maxiter/tolerance
    with mock.patch("pyhf.set_backend") as mock_backend:
        fit.significance(model, data)
    assert mock_backend.call_count == 2
    # setting to minuit
    assert mock_backend.call_args_list[0][0][1].name == "minuit"
    assert mock_backend.call_args_list[0][0][1].verbose == 1
    assert mock_backend.call_args_list[0][0][1].strategy is None
    assert mock_backend.call_args_list[0][0][1].maxiter is None
    assert mock_backend.call_args_list[0][0][1].tolerance is None
    assert mock_backend.call_args_list[0][1] == {}
    # resetting back to scipy
    assert mock_backend.call_args_list[1][0][1].name == "scipy"

    # custom strategy/maxiter/tolerance
    with mock.patch("pyhf.set_backend") as mock_backend:
        fit.significance(model, data, strategy=2, maxiter=100, tolerance=0.01)
    assert mock_backend.call_count == 2
    # setting to minuit
    assert mock_backend.call_args_list[0][0][1].name == "minuit"
    assert mock_backend.call_args_list[0][0][1].verbose == 1
    assert mock_backend.call_args_list[0][0][1].strategy == 2
    assert mock_backend.call_args_list[0][0][1].maxiter == 100
    assert mock_backend.call_args_list[0][0][1].tolerance == 0.01
    assert mock_backend.call_args_list[0][1] == {}
    # resetting back to scipy
    assert mock_backend.call_args_list[1][0][1].name == "scipy"



"""Visualizes fit models with matplotlib."""

import logging
import pathlib
from typing import Any, Dict, List, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import packaging.version

from cabinetry.visualize import utils


log = logging.getLogger(__name__)

# handling of matplotlib<3.6
if packaging.version.parse(mpl.__version__) < packaging.version.parse("3.6"):
    MPL_STYLE = "seaborn-colorblind"  # pragma: no cover
else:
    MPL_STYLE = "seaborn-v0_8-colorblind"


def data_mc(
    histogram_dict_list: List[Dict[str, Any]],
    total_model_unc: np.ndarray,
    bin_edges: np.ndarray,
    *,
    figure_path: Optional[pathlib.Path] = None,
    log_scale: Optional[bool] = None,
    log_scale_x: bool = False,
    label: str = "",
    colors: Optional[Dict[str, str]] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a data/MC histogram with uncertainty bands and ratio panel.

    Args:
        histogram_dict_list (List[Dict[str, Any]]): list of samples (with info stored in
            one dict per sample)
        total_model_unc (np.ndarray): total model uncertainty, if specified this is used
            instead of calculating it via sum in quadrature, defaults to None
        bin_edges (np.ndarray): bin edges of histogram
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        log_scale (Optional[bool], optional): whether to use a logarithmic vertical
            axis, defaults to None (automatically determine whether to use linear or log
            scale)
        log_scale_x (bool, optional): whether to use logarithmic horizontal axis,
            defaults to False
        label (str, optional): label written on the figure, defaults to ""
        colors (Optional[Dict[str, str]], optional): map of sample names and colors to
            use in plot, defaults to None (uses default colors)
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Raises:
        ValueError: when total model yield is negative in any bin

    Returns:
        matplotlib.figure.Figure: the data/MC figure
    """
    mc_histograms_yields = []
    mc_labels = []
    for h in histogram_dict_list:
        if h["isData"]:
            data_histogram_yields = h["yields"]
            data_histogram_stdev = np.sqrt(data_histogram_yields)
            data_label = h["label"]
        else:
            mc_histograms_yields.append(h["yields"])
            mc_labels.append(h["label"])

    mpl.style.use(MPL_STYLE)

    fig = plt.figure(figsize=(6, 6), layout="constrained")
    gs = fig.add_gridspec(nrows=2, ncols=1, hspace=0, height_ratios=[3, 1])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    # increase font sizes
    for item in (
        [ax1.yaxis.label, ax2.xaxis.label, ax2.yaxis.label]
        + ax1.get_yticklabels()
        + ax2.get_xticklabels()
        + ax2.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks on all axes
    for axis in [ax1.xaxis, ax1.yaxis, ax2.xaxis, ax2.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    # plot MC stacked together
    total_yield = np.zeros_like(mc_histograms_yields[0])
    bin_right_edges = bin_edges[1:]
    bin_left_edges = bin_edges[:-1]
    bin_width = bin_right_edges - bin_left_edges
    bin_centers = 0.5 * (bin_left_edges + bin_right_edges)
    # center data visually in bins if horizontal log scale is used
    bin_centers_data = (
        np.power(10, 0.5 * (np.log10(bin_left_edges * bin_right_edges)))
        if log_scale_x
        else bin_centers
    )
    mc_containers = []
    for mc_sample_yield, sample_label in zip(mc_histograms_yields, mc_labels):
        mc_container = ax1.bar(
            bin_centers,
            mc_sample_yield,
            width=bin_width,
            bottom=total_yield,
            color=colors[sample_label] if colors else None,
        )
        mc_containers.append(mc_container)

        # add a black line on top of each sample
        line_x = [y for y in bin_edges for _ in range(2)][1:-1]
        line_y = [y for y in (mc_sample_yield + total_yield) for _ in range(2)]
        ax1.plot(line_x, line_y, "-", color="black", linewidth=0.5)

        total_yield += mc_sample_yield

    # add total MC uncertainty
    mc_unc_container = ax1.bar(
        bin_centers,
        2 * total_model_unc,
        width=bin_width,
        bottom=total_yield - total_model_unc,
        fill=False,
        linewidth=0,
        edgecolor="gray",
        hatch=3 * "/",
    )

    # plot data
    data_container = ax1.errorbar(
        bin_centers_data,
        data_histogram_yields,
        yerr=data_histogram_stdev,
        fmt="o",
        color="k",
    )

    # ratio plot
    ax2.plot(
        [bin_left_edges[0], bin_right_edges[-1]],
        [1, 1],
        "--",
        color="black",
        linewidth=1,
    )  # reference line along y=1

    n_zero_pred = sum(total_yield == 0.0)  # number of bins with zero predicted yields
    if n_zero_pred > 0:
        log.warning(
            f"predicted yield is zero in {n_zero_pred} bin(s), excluded from ratio plot"
        )
    nonzero_model_yield = total_yield != 0.0

    if np.any(total_yield < 0.0):
        raise ValueError(
            f"{label} total model yield has negative bin(s): {total_yield.tolist()}"
        )

    # add uncertainty band around y=1
    rel_mc_unc = total_model_unc / total_yield
    # do not show band in bins where total model yield is 0
    ax2.bar(
        bin_centers[nonzero_model_yield],
        2 * rel_mc_unc[nonzero_model_yield],
        width=bin_width[nonzero_model_yield],
        bottom=1.0 - rel_mc_unc[nonzero_model_yield],
        fill=False,
        linewidth=0,
        edgecolor="gray",
        hatch=3 * "/",
    )

    # data in ratio plot
    data_model_ratio = data_histogram_yields / total_yield
    data_model_ratio_unc = data_histogram_stdev / total_yield
    # mask data in bins where total model yield is 0
    ax2.errorbar(
        bin_centers_data[nonzero_model_yield],
        data_model_ratio[nonzero_model_yield],
        yerr=data_model_ratio_unc[nonzero_model_yield],
        fmt="o",
        color="k",
    )

    # get the highest single bin yield, from the sum of MC or data
    y_max = max(np.amax(total_yield), np.amax(data_histogram_yields))
    # lowest model yield in single bin (not considering empty bins)
    y_min = np.amin(total_yield[np.nonzero(total_yield)])

    # use log scale if it is requested, otherwise determine scale setting:
    # if yields vary over more than 2 orders of magnitude, set y-axis to log scale
    if log_scale or (log_scale is None and (y_max / y_min) > 100):
        # log vertical axis scale and limits
        ax1.set_yscale("log")
        ax1.set_ylim([y_min / 10, y_max * 10])
        # add "_log" to the figure name if figure should be saved
        figure_path = utils._log_figure_path(figure_path)
    else:
        # do not use log scale
        ax1.set_ylim([0, y_max * 1.5])  # 50% headroom

    # log scale for horizontal axes
    if log_scale_x:
        ax1.set_xscale("log")
        ax2.set_xscale("log")

    # figure label (region name)
    at = mpl.offsetbox.AnchoredText(
        label,
        loc="upper left",
        frameon=False,
        prop={"fontsize": "large", "linespacing": 1.5},
    )
    ax1.add_artist(at)

    # MC contributions in inverse order, such that first legend entry corresponds to
    # the last (highest) contribution to the stack
    all_containers = mc_containers[::-1] + [mc_unc_container, data_container]
    all_labels = mc_labels[::-1] + ["Uncertainty", data_label]
    ax1.legend(
        all_containers, all_labels, frameon=False, fontsize="large", loc="upper right"
    )

    ax1.set_xlim(bin_edges[0], bin_edges[-1])
    ax1.set_ylabel("events")
    ax1.set_xticklabels([])
    ax1.set_xticklabels([], minor=True)
    ax1.tick_params(axis="both", which="major", pad=8)  # tick label - axis padding
    ax1.tick_params(direction="in", top=True, right=True, which="both")

    ax2.set_xlim(bin_edges[0], bin_edges[-1])
    ax2.set_ylim([0.5, 1.5])
    ax2.set_xlabel(histogram_dict_list[0]["variable"])
    ax2.set_ylabel("data / model")
    ax2.set_yticks([0.5, 0.75, 1.0, 1.25, 1.5])
    ax2.set_yticklabels([0.5, 0.75, 1.0, 1.25, ""])
    ax2.tick_params(axis="both", which="major", pad=8)
    ax2.tick_params(direction="in", top=True, right=True, which="both")

    utils._save_and_close(fig, figure_path, close_figure)
    return fig


def templates(
    nominal_histo: Dict[str, np.ndarray],
    up_histo_orig: Dict[str, np.ndarray],
    down_histo_orig: Dict[str, np.ndarray],
    up_histo_mod: Dict[str, np.ndarray],
    down_histo_mod: Dict[str, np.ndarray],
    bin_edges: np.ndarray,
    variable: str,
    *,
    figure_path: Optional[pathlib.Path] = None,
    label: str = "",
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a nominal template and the associated up/down variations.

    If a variation template is an empty dict, it is not drawn.

    Args:
        nominal_histo (Dict[str, np.ndarray]): the nominal template
        up_histo_orig (Dict[str, np.ndarray]): original "up" variation
        down_histo_orig (Dict[str, np.ndarray]): original "down" variation
        up_histo_mod (Dict[str, np.ndarray]): "up" variation after post-processing
        down_histo_mod (Dict[str, np.ndarray]): "down" variation after post-processing
        bin_edges (np.ndarray): bin edges of histogram
        variable (str): variable name for the horizontal axis
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        label (str, optional): label written on the figure, defaults to ""
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    Returns:
        matplotlib.figure.Figure: the template figure
    """
    bin_width = bin_edges[1:] - bin_edges[:-1]
    bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])

    mpl.style.use(MPL_STYLE)
    fig = plt.figure(figsize=(8, 6), layout="constrained")
    gs = fig.add_gridspec(nrows=2, ncols=1, hspace=0, height_ratios=[3, 1])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    # ratio plot line through unity and stat. uncertainty of nominal
    ax2.plot([bin_edges[0], bin_edges[-1]], [1, 1], "--", color="black", linewidth=1)
    rel_nominal_stat_unc = nominal_histo["stdev"] / nominal_histo["yields"]
    ax2.bar(
        bin_centers,
        2 * rel_nominal_stat_unc,
        width=bin_width,
        bottom=1 - rel_nominal_stat_unc,
        fill=False,
        linewidth=0,
        edgecolor="gray",
        hatch=3 * "/",
    )

    colors = ["black", "C0", "C1", "C0", "C1"]
    linestyles = ["-", ":", ":", "--", "--"]
    all_templates = [
        nominal_histo,
        up_histo_orig,
        down_histo_orig,
        up_histo_mod,
        down_histo_mod,
    ]
    template_labels = [
        "nominal",
        "up (original)",
        "down (original)",
        "up (modified)",
        "down (modified)",
    ]

    # x positions for lines drawn showing the template distributions
    line_x = [y for y in bin_edges for _ in range(2)][1:-1]

    neg_nom_bin = False  # negative bin(s) present in nominal histogram
    if np.any(nominal_histo["yields"] < 0.0):
        neg_nom_bin = True
        log.warning(
            f"{label} nominal histogram yield has negative bin(s): "
            f"{nominal_histo['yields'].tolist()}, taking absolute value for "
            "ratio plot uncertainty"
        )

    # draw templates
    for template, color, linestyle, template_label in zip(
        all_templates, colors, linestyles, template_labels
    ):
        if not template:
            # variation not defined
            continue

        # lines to show each template distribution
        line_y = [y for y in template["yields"] for _ in range(2)]

        ax1.plot(line_x, line_y, color=color, linestyle=linestyle, label=template_label)
        if template_label == "nominal":
            # band for stat. uncertainty of nominal prediction
            ax1.bar(
                bin_centers,
                2 * nominal_histo["stdev"],
                width=bin_width,
                bottom=nominal_histo["yields"] - nominal_histo["stdev"],
                fill=False,
                linewidth=0,
                edgecolor="gray",
                hatch=3 * "/",
            )
        else:
            # error bars for up/down variations
            ax1.errorbar(
                bin_centers,
                template["yields"],
                yerr=template["stdev"],
                fmt="none",
                color=color,
            )

            # ratio plot: variation / nominal
            template_ratio_plot = template["yields"] / nominal_histo["yields"]
            line_y = [y for y in template_ratio_plot for _ in range(2)]

            ax2.plot(line_x, line_y, color=color, linestyle=linestyle)
            ax2.errorbar(
                bin_centers,
                template_ratio_plot,
                yerr=template["stdev"] / np.abs(nominal_histo["yields"]),
                fmt="none",
                color=color,
            )

    # increase font sizes
    for item in (
        [ax1.yaxis.label, ax2.xaxis.label, ax2.yaxis.label]
        + ax1.get_yticklabels()
        + ax2.get_xticklabels()
        + ax2.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks on all axes
    for axis in [ax1.xaxis, ax1.yaxis, ax2.xaxis, ax2.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    # figure label (region, sample, systematic name)
    at = mpl.offsetbox.AnchoredText(
        label,
        loc="upper left",
        frameon=False,
        prop={"fontsize": "large", "linespacing": 1.5},
    )
    ax1.add_artist(at)

    ax1.legend(frameon=False, fontsize="large", loc="upper right")

    max_yield = max(max(template["yields"]) for template in all_templates if template)

    ax1.set_xlim([bin_edges[0], bin_edges[-1]])
    ax1.set_ylim([0, max_yield * 1.75])
    ax1.set_ylabel("events")
    ax1.set_xticklabels([])
    ax1.tick_params(axis="both", which="major", pad=8)  # tick label - axis padding
    ax1.tick_params(direction="in", top=True, right=True, which="both")

    ax2.set_xlim([bin_edges[0], bin_edges[-1]])
    ax2.set_ylim([0.5, 1.5])
    ax2.set_xlabel(variable)
    ax2.set_ylabel(f"variation / {'nominal' if not neg_nom_bin else 'abs(nominal)'}")
    ax2.set_yticks([0.5, 0.75, 1.0, 1.25, 1.5])
    ax2.set_yticklabels([0.5, 0.75, 1.0, 1.25, ""])
    ax2.tick_params(axis="both", which="major", pad=8)
    ax2.tick_params(direction="in", top=True, right=True, which="both")

    utils._save_and_close(fig, figure_path, close_figure)
    return fig


def modifier_grid(
    grid_list: List[np.ndarray],
    axis_labels: List[List[str]],
    category_map: Dict[int, str],
    figure_path: Optional[pathlib.Path] = None,
    close_figure: bool = False,
) -> mpl.figure.Figure:
    """Draws a grid of modifiers per channel, sample and parameter.

    Args:
        grid_list (List[np.ndarray]): list of 2d grids with modifier information
        axis_labels (List[List[str]]): list with axis labels for the three axes in order
            (first axis is grid label, second and third are axes per grid)
        category_map (Dict[int, str]): translation of integer values in grid to labels
        figure_path (Optional[pathlib.Path], optional): path where figure should be
            saved, or None to not save it, defaults to None
        close_figure (bool, optional): whether to close each figure immediately after
            saving it, defaults to False (enable when producing many figures to avoid
            memory issues, prevents rendering in notebooks)

    """
    # colors for categories from Okabe & Ito: https://jfly.uni-koeln.de/color/#pallet,
    # see also https://doi.org/10.1038/nmeth.1618
    cmap = mpl.colors.ListedColormap(
        [
            (1, 1, 1),  # white,
            (86 / 255, 180 / 255, 233 / 255),  # sky blue
            (240 / 255, 228 / 255, 66 / 255),  # yellow
            (0, 158 / 255, 115 / 255),  # blueish green
            (230 / 255, 159 / 255, 0),  # orange
            (85 / 255, 85 / 255, 85 / 255),  # dark gray
            (204 / 255, 121 / 255, 167 / 255),  # reddish purple
            (0, 114 / 255, 178 / 255),  # blue
            (213 / 255, 94 / 255, 0),  # vermilion
        ]
    ).reversed()
    norm = mpl.colors.BoundaryNorm(np.arange(-0.5, 9.5), cmap.N)

    # rough heuristics for figure size
    fig_width = 3 + len(axis_labels[2]) / 3.8
    fig_height = 3 + (len(axis_labels[0]) * len(axis_labels[1])) / 3.2
    fig, ax = plt.subplots(
        len(axis_labels[0]),
        sharex=True,
        figsize=(fig_width, fig_height),
        layout="constrained",
        squeeze=False,  # always return array of axes (even with a single subplot)
    )
    ax = ax.flatten()  # turn into 1d array

    for i_grid, grid_label in enumerate(axis_labels[0]):
        im = ax[i_grid].imshow(grid_list[i_grid], cmap=cmap, norm=norm)
        ax[i_grid].set_title(grid_label)
        ax[i_grid].set_xticks(np.arange(len(axis_labels[2])))  # sample or channel
        ax[i_grid].set_yticks(np.arange(len(axis_labels[1])))  # parameter
        ax[i_grid].set_xticklabels(axis_labels[2])
        ax[i_grid].set_yticklabels(axis_labels[1])
        for tick in ax[i_grid].get_xticklabels():
            tick.set_rotation(45)
            tick.set_horizontalalignment("right")

    # construct colobar with category labels
    formatter = plt.FuncFormatter(lambda val, _: category_map[val])
    _ = fig.colorbar(
        im, ax=ax.ravel().tolist(), ticks=np.arange(len(category_map)), format=formatter
    )

    utils._save_and_close(fig, figure_path, close_figure)
    return fig



from collections import defaultdict, namedtuple
import copy
import logging
import pathlib
from unittest import mock

import matplotlib.figure
import numpy as np
import pyhf
import pytest

from cabinetry import fit
from cabinetry import model_utils
from cabinetry import visualize


MockHistogram = namedtuple("MockHistogram", ["bins", "yields", "stdev"])


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("SR", "pre-fit"), "SR_prefit.pdf"),
        (("SR", "post-fit"), "SR_postfit.pdf"),
        (("SR 1", "prefit"), "SR-1_prefit.pdf"),
        (("SR 1", "postfit"), "SR-1_postfit.pdf"),
    ],
)
def test__figure_name(test_input, expected):
    assert visualize._figure_name(*test_input) == expected


def test__total_yield_uncertainty():
    stdev_list = [np.asarray([0.1, 0.2, 0.1]), np.asarray([0.3, 0.2, 0.1])]
    expected_uncertainties = [0.31622777, 0.28284271, 0.14142136]
    assert np.allclose(
        visualize._total_yield_uncertainty(stdev_list), expected_uncertainties
    )


@mock.patch("cabinetry.visualize._total_yield_uncertainty", return_value=[0.2])
@mock.patch(
    "cabinetry.visualize.plot_model.data_mc", return_value=matplotlib.figure.Figure()
)
@mock.patch(
    "cabinetry.histo.Histogram.from_config",
    return_value=MockHistogram([0.0, 1.0], [1.0], [0.1]),
)
def test_data_mc_from_histograms(mock_load, mock_draw, mock_stdev):
    config = {
        "General": {"HistogramFolder": "tmp_hist"},
        "Regions": [{"Name": "reg_1", "Variable": "x"}],
        "Samples": [{"Name": "sample_1"}, {"Name": "data", "Data": True}],
    }
    figure_folder = pathlib.Path("tmp")
    histogram_folder = pathlib.Path("tmp_hist")

    fig_dict_list = visualize.data_mc_from_histograms(
        config, figure_folder=figure_folder
    )
    assert len(fig_dict_list) == 1
    assert isinstance(fig_dict_list[0]["figure"], matplotlib.figure.Figure)
    assert fig_dict_list[0]["region"] == "reg_1"

    # the call_args_list contains calls (outer round brackets), first filled with
    # arguments (inner round brackets) and then keyword arguments
    assert mock_load.call_args_list == [
        (
            (
                histogram_folder,
                {"Name": "reg_1", "Variable": "x"},
                {"Name": "data", "Data": True},
                {},
            ),
            {"modified": True},
        ),
        (
            (
                histogram_folder,
                {"Name": "reg_1", "Variable": "x"},
                {"Name": "sample_1"},
                {},
            ),
            {"modified": True},
        ),
    ]
    assert mock_stdev.call_args_list == [(([[0.1]],), {})]
    assert mock_draw.call_args_list == [
        (
            (
                [
                    {"label": "data", "isData": True, "yields": [1.0], "variable": "x"},
                    {
                        "label": "sample_1",
                        "isData": False,
                        "yields": [1.0],
                        "variable": "x",
                    },
                ],
                [0.2],
                [0.0, 1.0],
            ),
            {
                "figure_path": figure_folder / "reg_1_prefit.pdf",
                "log_scale": None,
                "log_scale_x": False,
                "label": "reg_1\npre-fit",
                "colors": None,
                "close_figure": False,
            },
        )
    ]

    # custom log scale settings, close figure, do not save figure, channel specified,
    # custom colors
    _ = visualize.data_mc_from_histograms(
        config,
        figure_folder=figure_folder,
        log_scale=True,
        log_scale_x=True,
        channels=["reg_1"],
        colors={"sample_1": "red"},
        close_figure=True,
        save_figure=False,
    )
    assert mock_draw.call_args[1] == {
        "figure_path": None,
        "log_scale": True,
        "log_scale_x": True,
        "label": "reg_1\npre-fit",
        "colors": {"sample_1": "red"},
        "close_figure": True,
    }

    # no matching channels
    assert visualize.data_mc_from_histograms(config, channels="abc") == []

    # incomplete color specification
    with pytest.raises(
        ValueError,
        match="colors need to be provided for all samples, missing for {'sample_1'}",
    ):
        _ = visualize.data_mc_from_histograms(config, colors={})


@mock.patch(
    "cabinetry.visualize.plot_model.data_mc", return_value=matplotlib.figure.Figure()
)
@mock.patch("cabinetry.templates.builder._binning", return_value=np.asarray([1, 2]))
@mock.patch(
    "cabinetry.configuration.region_dict",
    side_effect=[
        {"Name": "region", "Variable": "x", "Binning": [1, 2]},
        {"Name": "region"},
    ],
)
@mock.patch(
    "cabinetry.model_utils._filter_channels",
    side_effect=[["Signal Region"], ["Signal Region"], ["Signal Region"], []],
)
@mock.patch("cabinetry.model_utils._data_per_channel", return_value=[[12.0]])
def test_data_mc(mock_data, mock_filter, mock_dict, mock_bins, mock_draw, example_spec):
    config = {"config": "abc"}
    figure_folder = "tmp"
    model = pyhf.Workspace(example_spec).model()
    model_pred = model_utils.ModelPrediction(
        model, [[[10.0]]], [[[0.3], [0.3]]], [[0.3, 0.3]], "pre-fit"
    )
    data = [12.0, 1.0]

    # pre-fit plot
    fig_dict_list = visualize.data_mc(
        model_pred, data, config=config, figure_folder=figure_folder
    )
    assert len(fig_dict_list) == 1
    assert isinstance(fig_dict_list[0]["figure"], matplotlib.figure.Figure)
    assert fig_dict_list[0]["region"] == "Signal Region"

    assert mock_data.call_args_list == [((model, data), {})]
    assert mock_filter.call_args_list == [((model, None), {})]
    assert mock_dict.call_args_list == [((config, "Signal Region"), {})]
    assert mock_bins.call_args_list == [
        (({"Name": "region", "Variable": "x", "Binning": [1, 2]},), {})
    ]

    expected_histograms = [
        {
            "label": "Signal",
            "isData": False,
            "yields": np.asarray([10.0]),
            "variable": "x",
        },
        {
            "label": "Data",
            "isData": True,
            "yields": np.asarray([12.0]),
            "variable": "x",
        },
    ]
    assert mock_draw.call_count == 1
    assert mock_draw.call_args_list[0][0][0] == expected_histograms
    assert np.allclose(mock_draw.call_args_list[0][0][1], np.asarray([0.3]))
    np.testing.assert_equal(mock_draw.call_args_list[0][0][2], np.asarray([1, 2]))
    assert mock_draw.call_args_list[0][1] == {
        "figure_path": pathlib.Path("tmp/Signal-Region_prefit.pdf"),
        "log_scale": None,
        "log_scale_x": False,
        "label": "Signal Region\npre-fit",
        "colors": None,
        "close_figure": False,
    }

    # post-fit plot (different label in model prediction), custom scale, close figure,
    # do not save figure, custom colors, histogram input mode: no binning or variable
    # specified (via side effect)
    model_pred = model_utils.ModelPrediction(
        model, [[[11.0]]], [[[0.2], [0.2]]], [[0.2, 0.2]], "post-fit"
    )
    _ = visualize.data_mc(
        model_pred,
        data,
        config=config,
        figure_folder=figure_folder,
        log_scale=False,
        colors={"Signal": "red"},
        close_figure=True,
        save_figure=False,
    )

    assert mock_draw.call_count == 2
    # yield at best-fit point is different from pre-fit
    assert np.allclose(mock_draw.call_args[0][0][0]["yields"], 11.0)
    assert np.allclose(mock_draw.call_args[0][1], np.asarray([0.2]))
    # observable defaults to "bin"
    assert mock_draw.call_args[0][0][0]["variable"] == "bin"
    assert mock_draw.call_args[0][0][1]["variable"] == "bin"
    # binning falls back to default
    np.testing.assert_equal(mock_draw.call_args[0][2], np.asarray([0, 1]))
    assert mock_draw.call_args[1] == {
        "figure_path": None,  # figure not saved
        "log_scale": False,
        "log_scale_x": False,
        "label": "Signal Region\npost-fit",
        "colors": {"Signal": "red"},
        "close_figure": True,
    }

    # no config specified, default variable name and bin edges
    _ = visualize.data_mc(model_pred, data)
    assert mock_draw.call_args[0][0][0]["variable"] == "bin"
    assert mock_draw.call_args[0][0][1]["variable"] == "bin"
    assert mock_draw.call_args[0][0][1]["yields"] == np.asarray(data[:1])
    np.testing.assert_equal(mock_draw.call_args[0][2], np.asarray([0, 1]))

    # no matching channels (via side_effect)
    assert visualize.data_mc(model_pred, data, channels="abc") is None
    assert mock_filter.call_args == ((model, "abc"), {})
    assert mock_draw.call_count == 3  # no new call

    # incomplete color specification
    with pytest.raises(
        ValueError,
        match="colors need to be provided for all samples, missing for {'Signal'}",
    ):
        _ = visualize.data_mc(model_pred, data, colors={})


@mock.patch(
    "cabinetry.histo.Histogram.from_path",
    side_effect=[
        MockHistogram([0.0, 1.0], [2.0], [0.2]),
        MockHistogram([0.0, 1.0], [3.0], [0.3]),
        MockHistogram([0.0, 1.0], [4.0], [0.4]),
        MockHistogram([0.0, 1.0], [5.0], [0.5]),
    ]
    * 3,
)
@mock.patch(
    "cabinetry.histo.Histogram.from_config",
    return_value=MockHistogram([0.0, 1.0], [1.0], [0.1]),
)
@mock.patch(
    "cabinetry.visualize.plot_model.templates", return_value=matplotlib.figure.Figure()
)
def test_templates(mock_draw, mock_histo_config, mock_histo_path, tmp_path):
    # the side effects are repeated for the patched Histogram.from_path
    # to check all relevant behavior (including the unknown backend check)
    nominal_path = tmp_path / "region_sample_Nominal_modified.npz"
    up_path = tmp_path / "region_sample_sys-name_Up_modified.npz"
    down_path = tmp_path / "region_sample_sys-name_Down_modified.npz"
    region = {"Name": "region", "Variable": "x"}
    sample = {"Name": "sample"}
    config = {
        "General": {"HistogramFolder": tmp_path},
        "Regions": [region],
        "Samples": [sample, {"Name": "data", "Data": True}],
        "Systematics": [{"Name": "sys name"}],
    }

    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "templates/region_sample_sys-name.pdf"

    # add fake histograms for glob
    nominal_path.touch()
    up_path.touch()
    down_path.touch()

    # also add a file that matches pattern but is not needed
    (tmp_path / "region_sample_sys-name_unknown_modified.npz").touch()

    fig_dict_list = visualize.templates(config, figure_folder=folder_path)
    assert len(fig_dict_list) == 1
    assert isinstance(fig_dict_list[0]["figure"], matplotlib.figure.Figure)
    assert fig_dict_list[0]["region"] == "region"
    assert fig_dict_list[0]["sample"] == "sample"
    assert fig_dict_list[0]["systematic"] == "sys name"

    # nominal histogram loading
    assert mock_histo_config.call_args_list == [((tmp_path, region, sample, {}), {})]
    # variation histograms
    down_path_orig = pathlib.Path(str(down_path).replace("_modified", ""))
    up_path_orig = pathlib.Path(str(up_path).replace("_modified", ""))
    assert mock_histo_path.call_args_list == [
        ((down_path_orig,), {}),
        ((down_path,), {}),
        ((up_path_orig,), {}),
        ((up_path,), {}),
    ]

    nominal = {"yields": [1.0], "stdev": [0.1]}
    up_orig = {"yields": [4.0], "stdev": [0.4]}
    down_orig = {"yields": [2.0], "stdev": [0.2]}
    up_mod = {"yields": [5.0], "stdev": [0.5]}
    down_mod = {"yields": [3.0], "stdev": [0.3]}
    bins = [0.0, 1.0]
    assert mock_draw.call_args_list == [
        (
            (nominal, up_orig, down_orig, up_mod, down_mod, bins, "x"),
            {
                "figure_path": figure_path,
                "label": "region: region\nsample: sample\nsystematic: sys name",
                "close_figure": False,
            },
        )
    ]

    # close figure, do not save figure, and remove variable information from config
    # (simulating histogram inputs), so variable defaults to "observable"
    histo_config = copy.deepcopy(config)
    histo_config["Regions"] = [{"Name": "region"}]
    _ = visualize.templates(
        histo_config, figure_folder=folder_path, close_figure=True, save_figure=False
    )
    assert mock_draw.call_args == (
        (nominal, up_orig, down_orig, up_mod, down_mod, bins, "observable"),
        {
            "figure_path": None,
            "label": "region: region\nsample: sample\nsystematic: sys name",
            "close_figure": True,
        },
    )

    # remove files for variation histograms
    up_path.unlink()
    down_path.unlink()

    assert mock_draw.call_count == 2  # two calls so far
    _ = visualize.templates(config, figure_folder=folder_path)
    assert mock_draw.call_count == 2  # no new call, since no variations found

    # no systematics in config
    config = {
        "General": {"HistogramFolder": tmp_path},
        "Regions": [region],
        "Samples": [sample, {"Name": "data", "Data": True}],
    }
    _ = visualize.templates(config, figure_folder=folder_path)
    assert mock_draw.call_count == 2  # no systematics, so no new calls


@mock.patch(
    "cabinetry.visualize.plot_result.correlation_matrix",
    return_value=matplotlib.figure.Figure(),
)
def test_correlation_matrix(mock_draw):
    corr_mat = np.asarray([[1.0, 0.2, 0.1], [0.2, 1.0, 0.1], [0.1, 0.1, 1.0]])
    corr_mat_pruned = np.asarray([[1.0, 0.2], [0.2, 1.0]])
    labels = ["a", "b", "c"]
    labels_pruned = ["a", "b"]
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "correlation_matrix.pdf"
    fit_results = fit.FitResults(np.empty(0), np.empty(0), labels, corr_mat, 1.0)

    # pruning with threshold
    fig = visualize.correlation_matrix(
        fit_results, figure_folder=folder_path, pruning_threshold=0.15
    )
    assert isinstance(fig, matplotlib.figure.Figure)

    mock_draw.assert_called_once()
    assert np.allclose(mock_draw.call_args[0][0], corr_mat_pruned)
    assert np.any(
        [mock_draw.call_args[0][1][i] == labels[i] for i in range(len(labels_pruned))]
    )
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # pruning of fixed parameter (all zeros in correlation matrix row/column), do not
    # close figure, do not save
    corr_mat_fixed = np.asarray([[1.0, 0.2, 0.0], [0.2, 1.0, 0.0], [0.0, 0.0, 0.0]])
    fit_results_fixed = fit.FitResults(
        np.empty(0), np.empty(0), labels, corr_mat_fixed, 1.0
    )
    _ = visualize.correlation_matrix(
        fit_results_fixed,
        figure_folder=folder_path,
        close_figure=False,
        save_figure=False,
    )
    assert np.allclose(mock_draw.call_args_list[1][0][0], corr_mat_pruned)
    assert np.any(
        [
            mock_draw.call_args_list[1][0][1][i] == labels[i]
            for i in range(len(labels_pruned))
        ]
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.pulls", return_value=matplotlib.figure.Figure()
)
def test_pulls(mock_draw):
    bestfit = np.asarray([0.8, 1.0, 1.05, 1.1])
    uncertainty = np.asarray([0.9, 1.0, 0.03, 0.7])
    labels = ["a", "b", "staterror_region[0]", "c"]
    exclude = ["a"]
    folder_path = "tmp"
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 1.0)

    filtered_bestfit = np.asarray([1.0, 1.1])
    filtered_uncertainty = np.asarray([1.0, 0.7])
    filtered_labels = np.asarray(["b", "c"])
    figure_path = pathlib.Path(folder_path) / "pulls.pdf"

    # with filtering
    fig = visualize.pulls(fit_results, figure_folder=folder_path, exclude=exclude)
    assert isinstance(fig, matplotlib.figure.Figure)

    mock_draw.assert_called_once()
    assert np.allclose(mock_draw.call_args[0][0], filtered_bestfit)
    assert np.allclose(mock_draw.call_args[0][1], filtered_uncertainty)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == filtered_labels[i]
            for i in range(len(filtered_labels))
        ]
    )
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # filtering single parameter instead of list
    _ = visualize.pulls(fit_results, figure_folder=folder_path, exclude=exclude[0])

    assert np.allclose(mock_draw.call_args[0][0], filtered_bestfit)
    assert np.allclose(mock_draw.call_args[0][1], filtered_uncertainty)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == filtered_labels[i]
            for i in range(len(filtered_labels))
        ]
    )

    # without filtering via list, but with staterror removal, fixed parameter removal,
    # not closing figure, not saving
    fit_results.uncertainty[0] = 0.0

    bestfit_expected = np.asarray([1.0, 1.1])
    uncertainty_expected = np.asarray([1.0, 0.7])
    labels_expected = ["b", "c"]
    visualize.pulls(
        fit_results, figure_folder=folder_path, close_figure=False, save_figure=False
    )

    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected)
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == labels_expected[i]
            for i in range(len(labels_expected))
        ]
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.ranking", return_value=matplotlib.figure.Figure()
)
def test_ranking(mock_draw):
    bestfit = np.asarray([1.2, 0.1])
    uncertainty = np.asarray([0.2, 0.8])
    labels = ["staterror_a", "modeling"]
    impact_prefit_up = np.asarray([0.1, 0.5])
    impact_prefit_down = np.asarray([-0.2, -0.4])
    impact_postfit_up = np.asarray([0.1, 0.4])
    impact_postfit_down = np.asarray([-0.2, -0.3])
    ranking_results = fit.RankingResults(
        bestfit,
        uncertainty,
        labels,
        impact_prefit_up,
        impact_prefit_down,
        impact_postfit_up,
        impact_postfit_down,
    )
    folder_path = "tmp"

    figure_path = pathlib.Path(folder_path) / "ranking.pdf"
    bestfit_expected = np.asarray([0.1, 1.2])
    uncertainty_expected = np.asarray([0.8, 0.2])
    labels_expected = ["modeling", "staterror_a"]

    fig = visualize.ranking(ranking_results, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected)
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected)
    for i_lab, label in enumerate(mock_draw.call_args[0][2]):
        assert label == labels_expected[i_lab]
    assert np.allclose(mock_draw.call_args[0][3], impact_prefit_up[::-1])
    assert np.allclose(mock_draw.call_args[0][4], impact_prefit_down[::-1])
    assert np.allclose(mock_draw.call_args[0][5], impact_postfit_up[::-1])
    assert np.allclose(mock_draw.call_args[0][6], impact_postfit_down[::-1])
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # maximum parameter amount specified, do not close figure, do not save figure
    _ = visualize.ranking(
        ranking_results,
        figure_folder=folder_path,
        max_pars=1,
        close_figure=False,
        save_figure=False,
    )
    assert mock_draw.call_count == 2
    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected[0])
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected[0])
    assert mock_draw.call_args[0][2] == labels_expected[0]
    assert np.allclose(mock_draw.call_args[0][3], impact_prefit_up[1])
    assert np.allclose(mock_draw.call_args[0][4], impact_prefit_down[1])
    assert np.allclose(mock_draw.call_args[0][5], impact_postfit_up[1])
    assert np.allclose(mock_draw.call_args[0][6], impact_postfit_down[1])
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.scan", return_value=matplotlib.figure.Figure()
)
def test_scan(mock_draw):
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "scan_a_0.pdf"

    par_name = "a[0]"
    par_mle = 1.5
    par_unc = 0.2
    par_vals = np.asarray([1.3, 1.5, 1.7])
    par_nlls = np.asarray([0.9, 0.0, 1.1])
    scan_results = fit.ScanResults(par_name, par_mle, par_unc, par_vals, par_nlls)

    fig = visualize.scan(scan_results, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_draw.call_count == 1
    assert mock_draw.call_args[0][0] == par_name
    assert mock_draw.call_args[0][1] == par_mle
    assert mock_draw.call_args[0][2] == par_unc
    assert np.allclose(mock_draw.call_args[0][3], par_vals)
    assert np.allclose(mock_draw.call_args[0][4], par_nlls)
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # do not close figure, do not save figure
    _ = visualize.scan(
        scan_results, figure_folder=folder_path, close_figure=False, save_figure=False
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.limit", return_value=matplotlib.figure.Figure()
)
def test_limit(mock_draw):
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "limit.pdf"

    observed_CLs = np.asarray([0.75, 0.32, 0.02])
    expected_CLs = np.asarray([[0.1, 0.2, 0.3, 0.4, 0.5] for _ in range(3)])
    poi_values = np.asarray([0, 1, 2])
    confidence_level = 0.90
    limit_results = fit.LimitResults(
        3.0, np.empty(5), observed_CLs, expected_CLs, poi_values, confidence_level
    )

    fig = visualize.limit(limit_results, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], limit_results.observed_CLs)
    assert np.allclose(mock_draw.call_args[0][1], limit_results.expected_CLs)
    assert np.allclose(mock_draw.call_args[0][2], limit_results.poi_values)
    assert np.allclose(mock_draw.call_args[0][3], 1 - limit_results.confidence_level)
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # do not close figure, do not save figure
    _ = visualize.limit(
        limit_results, figure_folder=folder_path, close_figure=False, save_figure=False
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_model.modifier_grid",
    return_value=matplotlib.figure.Figure(),
)
@mock.patch(
    "cabinetry.model_utils._modifier_map",
    return_value=defaultdict(
        list, {("Signal Region", "Signal", "Signal strength"): ["normfactor"]}
    ),
)
def test_modifier_grid(mock_map, mock_draw, example_spec, caplog):
    caplog.set_level(logging.DEBUG)
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "modifier_grid.pdf"

    model = pyhf.Workspace(example_spec).model()
    # model contains a staterror, but remove that from the map via mocked _modifier_map
    # return to capture the effect of fields in the grid without modifiers

    fig = visualize.modifier_grid(model, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_map.call_count == 1
    assert mock_map.call_args[0][0].spec == model.spec
    assert mock_map.call_args[1] == {}

    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], [np.asarray([[0.0, 8.0]])])
    assert mock_draw.call_args[0][1] == [
        ["Signal Region"],
        ["Signal"],
        ["Signal strength", "staterror_Signal-Region"],
    ]
    assert mock_draw.call_args[0][2] == {
        0: "normfactor",
        1: "shapefactor",
        2: "shapesys",
        3: "lumi",
        4: "staterror",
        5: "normsys + histosys",
        6: "histosys",
        7: "normsys",
        8: "none",
    }
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # do not close figure, do not save figure, split by sample
    _ = visualize.modifier_grid(
        model,
        figure_folder=folder_path,
        split_by_sample=True,
        close_figure=False,
        save_figure=False,
    )
    assert np.allclose(mock_draw.call_args[0][0], [np.asarray([[0.0, 8.0]])])
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}
    assert mock_draw.call_args[0][1] == [
        ["Signal"],
        ["Signal Region"],
        ["Signal strength", "staterror_Signal-Region"],
    ]  # order changed due to split_by_sample
    caplog.clear()

    # unknown modifier combination: patch modifier map to return normfactor+staterror
    model = mock.MagicMock()
    model.config.channels = ["Signal Region"]
    model.config.samples = ["Signal"]
    model.config.par_order = ["Signal strength"]
    with mock.patch(
        "cabinetry.model_utils._modifier_map",
        return_value=defaultdict(
            list,
            {
                ("Signal Region", "Signal", "Signal strength"): [
                    "normfactor",
                    "staterror",
                ]
            },
        ),
    ):
        with pytest.raises(KeyError, match=r"staterror \+ normfactor"):
            _ = visualize.modifier_grid(model)

    assert "modifiers for Signal Region, Signal, Signal strength not supported" not in [
        rec.message for rec in caplog.records
    ]



"""High-level entry point to create, collect and post-process template histograms."""

import logging
import pathlib
from typing import Any, Dict, Optional

from cabinetry import route
from cabinetry.templates import builder
from cabinetry.templates import collector
from cabinetry.templates import postprocessor


log = logging.getLogger(__name__)


def build(
    config: Dict[str, Any],
    *,
    method: str = "uproot",
    router: Optional[route.Router] = None,
) -> None:
    """Produces all required histograms specified by the configuration file.

    Inputs to the histogram production are ntuples containing columnar data. Uses either
    a default method specified via ``method``, or a custom user-defined override through
    ``router``.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        method (str, optional): backend to use for histogram production, defaults to
            "uproot"
        router (Optional[route.Router], optional): instance of cabinetry.route.Router
            that contains user-defined overrides, defaults to None
    """
    # create an instance of the class doing the template building
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    general_path = config["General"]["InputPath"]
    template_builder = builder._Builder(histogram_folder, general_path, method)

    match_func: Optional[route.MatchFunc] = None
    if router is not None:
        # specify the wrapper for user-defined functions
        router.template_builder_wrapper = template_builder._wrap_custom_template_builder
        # get a function that can be queried to return a user-defined template builder
        match_func = router._find_template_builder_match

    route.apply_to_all_templates(
        config, template_builder._create_histogram, match_func=match_func
    )


def collect(config: Dict[str, Any], *, method: str = "uproot") -> None:
    """Collects all required histograms specified by the configuration file.

    Histograms must already exist, and this collects and saves them in the format used
    for further processing. If no default for VariationPath is specified in the general
    settings, it defaults to an empty string.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        method (str, optional): backend to use for histogram production, defaults to
            "uproot"
    """
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    general_path = config["General"]["InputPath"]
    variation_path = config["General"].get("VariationPath", None)
    if variation_path is None:
        # default to empty string and emit warning
        variation_path = ""
        log.info('no VariationPath specified in general settings, defaulting to ""')
    processor = collector._collector(
        histogram_folder, general_path, variation_path, method
    )
    route.apply_to_all_templates(config, processor)


def postprocess(config: Dict[str, Any]) -> None:
    """Applies postprocessing to all histograms.

    Args:
        config (Dict[str, Any]): cabinetry configuration
    """
    histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])
    processor = postprocessor._postprocessor(histogram_folder)
    route.apply_to_all_templates(config, processor)



"""Creates yield tables."""

import logging
import pathlib
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pyhf
import tabulate

from cabinetry import model_utils


log = logging.getLogger(__name__)


def _header_name(channel_name: str, i_bin: int, *, unique: bool = True) -> str:
    """Constructs the header name for a column in a yield table.

    There are two modes: by default the names are unique (to be used as keys). With
    ``unique=False``, the channel names are skipped for bins beyond the first bin (for
    less redundant output).

    Args:
        channel_name (str): name of the channel (phase space region)
        i_bin (int): index of bin in channel
        unique (bool, optional): whether to return a unique key, defaults to True

    Returns:
        str: the header name to be used for the column
    """
    if i_bin == 0 or unique:
        header_name = f"{channel_name}\nbin {i_bin+1}"
    else:
        header_name = f"\nbin {i_bin+1}"
    return header_name


def _save_tables(
    table_dict: Dict[str, List[Dict[str, Any]]],
    table_folder: pathlib.Path,
    table_label: str,
    table_format: str,
) -> None:
    """Saves yield tables in a specific format in a given folder.

    Newlines are removed from table headers for table formats that do not support them.

    Args:
        table_dict (Dict[str, List[Dict[str, Any]]]): dictionary with tables to save
        table_folder (pathlib.Path): path to the folder to save tables in
        table_label (str): label for tables to include in filenames
        table_format (str): format in which to save the tables
    """
    if table_format in ["plain", "simple", "tsv"]:
        save_suffix = "txt"
    elif table_format == "latex":
        save_suffix = "tex"
    else:
        save_suffix = table_format

    table_folder.mkdir(parents=True, exist_ok=True)

    for table_type in table_dict.keys():
        table_path = table_folder / f"{table_type}_{table_label}.{save_suffix}"

        if table_type == "yields_per_bin" and table_format in ["html", "latex", "tsv"]:
            # replace newlines in table headers for formats that do not support them
            # (newlines exist between channel name and bin number)
            for i_sample, sample in enumerate(table_dict[table_type]):
                table_dict[table_type][i_sample] = {
                    bin_name.replace("\n", ", "): sample_yield
                    for bin_name, sample_yield in sample.items()
                }

        table_str = (
            tabulate.tabulate(
                table_dict[table_type], headers="keys", tablefmt=table_format
            )
            + "\n"  # tabulate does not add a newline at the end
        )
        log.info(f"saving table as {table_path}")
        table_path.write_text(table_str)


def _yields_per_bin(
    model: pyhf.pdf.Model,
    model_yields: List[List[List[float]]],
    total_stdev_model: List[List[List[float]]],
    data: List[List[float]],
    channels: List[str],
    label: str,
) -> List[Dict[str, Any]]:
    """Outputs and returns a yield table with predicted and observed yields per bin.

    Args:
        model (pyhf.pdf.Model): the model which the table corresponds to
        model_yields (List[List[List[float]]]): yields per channel, sample, and bin
        total_stdev_model (List[List[List[float]]]): total model standard deviation per
            channel, sample and bin
        data (List[List[float]]): data yield per channel and bin
        channels (List[str]): names of channels to use
        label (str): label for model prediction to include in log

    Returns:
        List[Dict[str, Any]]: yield table for use with the ``tabulate`` package
    """
    table = []  # table containing all yields
    headers = {}  # headers with nicer formatting for output

    # indices of included channels
    channel_indices = [model.config.channels.index(ch) for ch in channels]

    # rows for each individual sample
    for i_sam, sample_name in enumerate(model.config.samples):
        sample_dict = {"sample": sample_name}  # one dict per sample
        for i_chan, channel_name in zip(channel_indices, channels):
            for i_bin in range(model.config.channel_nbins[channel_name]):
                sample_dict.update(
                    {
                        _header_name(
                            channel_name, i_bin
                        ): f"{model_yields[i_chan][i_sam][i_bin]:.2f} "
                        f"\u00B1 {total_stdev_model[i_chan][i_sam][i_bin]:.2f}"
                    }
                )
        table.append(sample_dict)

    # dicts for total model prediction and data
    total_dict = {"sample": "total"}
    data_dict = {"sample": "data"}
    for i_chan, channel_name in zip(channel_indices, channels):
        total_model = np.sum(model_yields[i_chan], axis=0)  # sum over samples
        for i_bin in range(model.config.channel_nbins[channel_name]):
            header_name = _header_name(channel_name, i_bin)
            headers.update(
                {header_name: _header_name(channel_name, i_bin, unique=False)}
            )
            total_dict.update(
                {
                    header_name: (
                        f"{total_model[i_bin]:.2f} "
                        f"\u00B1 {total_stdev_model[i_chan][-1][i_bin]:.2f}"
                    )
                }
            )
            data_dict.update({header_name: f"{data[i_chan][i_bin]:.2f}"})
    table += [total_dict, data_dict]

    log.info(
        f"yields per bin for {label} model prediction:\n"
        + tabulate.tabulate(table, headers=headers, tablefmt="fancy_grid")
    )
    return table


def _yields_per_channel(
    model: pyhf.pdf.Model,
    model_yields: List[List[float]],
    total_stdev_model: List[List[float]],
    data: List[float],
    channels: List[str],
    label: str,
) -> List[Dict[str, Any]]:
    """Outputs and returns a yield table with predicted and observed yields per channel.

    Args:
        model (pyhf.pdf.Model): the model which the table corresponds to
        model_yields (List[List[float]]): yields per channel and sample
        total_stdev_model (List[List[float]]): total model standard deviation per
            channel and sample
        data (List[float]): data yield per channel
        channels (List[str]): names of channels to use
        label (str): label for model prediction to include in log

    Returns:
        List[Dict[str, Any]]: yield table for use with the ``tabulate`` package
    """
    table = []  # table containing all yields

    # indices of included channels
    channel_indices = [model.config.channels.index(ch) for ch in channels]

    # rows for each individual sample
    for i_sam, sample_name in enumerate(model.config.samples):
        sample_dict = {"sample": sample_name}  # one dict per sample
        for i_chan, channel_name in zip(channel_indices, channels):
            sample_dict.update(
                {
                    channel_name: f"{model_yields[i_chan][i_sam]:.2f} "
                    f"\u00B1 {total_stdev_model[i_chan][i_sam]:.2f}"
                }
            )
        table.append(sample_dict)

    # dicts for total model prediction and data
    total_dict = {"sample": "total"}
    data_dict = {"sample": "data"}
    for i_chan, channel_name in zip(channel_indices, channels):
        total_model = np.sum(model_yields[i_chan], axis=0)  # sum over samples
        total_dict.update(
            {
                channel_name: f"{total_model:.2f} "
                f"\u00B1 {total_stdev_model[i_chan][-1]:.2f}"
            }
        )
        data_dict.update({channel_name: f"{data[i_chan]:.2f}"})
    table += [total_dict, data_dict]

    log.info(
        f"yields per channel for {label} model prediction:\n"
        + tabulate.tabulate(table, headers="keys", tablefmt="fancy_grid")
    )
    return table


def yields(
    model_prediction: model_utils.ModelPrediction,
    data: List[float],
    *,
    channels: Optional[Union[str, List[str]]] = None,
    per_bin: bool = True,
    per_channel: bool = False,
    table_folder: Union[str, pathlib.Path] = "tables",
    table_format: str = "simple",
    save_tables: bool = True,
) -> Dict[str, List[Dict[str, Any]]]:
    """Generates yield tables, showing model prediction and data.

    Channels can be filtered via the optional ``channels`` argument. Either yields per
    bin, or yields per channel, or both can be shown.

    Args:
        model_prediction (model_utils.ModelPrediction): model prediction to show in
            table
        data (List[float]): data to include in table, can either include auxdata (the
            auxdata is then stripped internally) or only observed yields
        channels (Optional[Union[str, List[str]]], optional): name of channel to show,
            or list of names to include, defaults to None (uses all channels)
        per_bin (bool, optional): whether to show a table with yields per bin, defaults
            to True
        per_channel (bool, optional): whether to show a table with yields per channel,
            defaults to False
        table_folder (Union[str, pathlib.Path], optional): path to the folder to save
            tables in, defaults to "tables"
        table_format (str, optional): format in which to save the tables, can be any of
            the formats ``tabulate`` supports (e.g. html, latex, plain, simple, tsv),
            defaults to "simple"
        save_tables (bool, optional): whether to save tables, defaults to True

    Returns:
        Dict[str, List[Dict[str, Any]]]: dictionary with yield tables for use with the
        ``tabulate`` package
    """
    table_dict: Dict[str, List[Dict[str, Any]]] = {}

    if not (per_bin or per_channel):
        log.warning(
            "requested neither yields per bin nor per channel, no table produced"
        )
        return table_dict

    # strip off auxdata (if needed) and obtain data indexed by channel (and bin)
    data_yields = model_utils._data_per_channel(model_prediction.model, data)

    # channels to include in table, with optional filtering applied
    filtered_channels = model_utils._filter_channels(model_prediction.model, channels)

    if filtered_channels == []:
        # nothing to include in tables, warning already raised via _filter_channels
        return table_dict

    if per_bin:
        # yield table with yields per bin
        per_bin_table = _yields_per_bin(
            model_prediction.model,
            model_prediction.model_yields,
            model_prediction.total_stdev_model_bins,
            data_yields,
            filtered_channels,
            model_prediction.label,
        )
        table_dict.update({"yields_per_bin": per_bin_table})

    if per_channel:
        # yields per channel
        n_channels = len(model_prediction.model.config.channels)
        model_yields_per_channel = [
            np.sum(model_prediction.model_yields[i_chan], axis=-1).tolist()
            for i_chan in range(n_channels)
        ]

        data_per_channel = [sum(d) for d in data_yields]
        per_channel_table = _yields_per_channel(
            model_prediction.model,
            model_yields_per_channel,
            model_prediction.total_stdev_model_channels,
            data_per_channel,
            filtered_channels,
            model_prediction.label,
        )
        table_dict.update({"yields_per_channel": per_channel_table})

    if save_tables:
        # save tables to file
        _save_tables(
            table_dict, pathlib.Path(table_folder), model_prediction.label, table_format
        )

    return table_dict



"""The cabinetry library."""

import logging
from typing import List

import cabinetry.configuration as configuration  # noqa: F401
import cabinetry.fit as fit  # noqa: F401
import cabinetry.histo as histo  # noqa: F401
import cabinetry.model_utils as model_utils  # noqa: F401
import cabinetry.route as route  # noqa: F401
import cabinetry.smooth as smooth  # noqa: F401
import cabinetry.tabulate as tabulate  # noqa: F401
import cabinetry.templates as templates  # noqa: F401
import cabinetry.visualize as visualize  # noqa: F401
import cabinetry.workspace as workspace  # noqa: F401


__all__ = [
    "__version__",
    "set_logging",
    "configuration",
    "fit",
    "histo",
    "model_utils",
    "route",
    "smooth",
    "tabulate",
    "templates",
    "visualize",
    "workspace",
]


def __dir__() -> List[str]:
    return __all__


__version__ = "0.6.0"


def set_logging() -> None:
    """Sets up customized and verbose logging output.

    Logging can be alternatively customized with the Python ``logging`` module directly.
    """
    logging.basicConfig(format="%(levelname)s - %(name)s - %(message)s")
    logging.getLogger("cabinetry").setLevel(logging.DEBUG)
    logging.getLogger("pyhf").setLevel(logging.INFO)



import numpy as np

from cabinetry.contrib import histogram_reader


def test_with_uproot(tmp_path, utils):
    fname = tmp_path / "test.root"
    histname = "h1"
    bins = np.asarray([1, 2, 3])
    yields = np.asarray([1.0, 2.0])
    stdev = np.asarray([0.1, 0.2])
    # create something to read
    utils.create_histogram(fname, histname, bins, yields, stdev)

    hist = histogram_reader.with_uproot(f"{fname}:{histname}")
    assert np.allclose(hist.axes[0].edges, bins)
    assert np.allclose(hist.values(), yields)
    assert np.allclose(np.sqrt(hist.variances()), stdev)



import functools
import logging
from unittest import mock

import boost_histogram as bh
import numpy as np
import pytest

from cabinetry import route


class ProcessorExamples:
    @staticmethod
    def get_example_template_builder():
        def example_template_builder(
            reg: dict, sam: dict, sys: dict, tem: str
        ) -> bh.Histogram:
            hist = bh.Histogram(bh.axis.Variable([0, 1]), storage=bh.storage.Weight())
            yields = np.asarray([2])
            stdev = np.asarray([0.1])
            hist[...] = np.stack([yields, stdev**2], axis=-1)
            return hist

        return example_template_builder


@pytest.fixture
def processor_examples():
    return ProcessorExamples


def test_Router():
    router = route.Router()
    assert router.template_builders == []
    assert router.template_builder_wrapper is None


def test_Router__register_processor(processor_examples):
    example_router = route.Router()
    example_template_builder = processor_examples.get_example_template_builder()

    example_router._register_processor(
        example_router.template_builders,
        region_name="reg",
        sample_name="signal",
        systematic_name="sys",
        template="Up",
    )(example_template_builder)

    assert example_router.template_builders == [
        {
            "region": "reg",
            "sample": "signal",
            "systematic": "sys",
            "template": "Up",
            "name": "example_template_builder",
            "func": example_template_builder,
        }
    ]


def test_Router_register_template_builder(processor_examples):
    example_router = route.Router()
    example_template_builder = processor_examples.get_example_template_builder()

    example_router.register_template_builder(
        region_name="reg", sample_name="signal", systematic_name="sys", template="Up"
    )(example_template_builder)

    assert example_router.template_builders == [
        {
            "region": "reg",
            "sample": "signal",
            "systematic": "sys",
            "template": "Up",
            "name": "example_template_builder",
            "func": example_template_builder,
        }
    ]

    # defaults for arguments
    example_router.register_template_builder()(example_template_builder)
    assert example_router.template_builders[1] == {
        "region": "*",
        "sample": "*",
        "systematic": "*",
        "template": "*",
        "name": "example_template_builder",
        "func": example_template_builder,
    }


def test_Router__find_match(processor_examples, caplog):
    caplog.set_level(logging.DEBUG)
    example_router = route.Router()
    example_template_builder_1 = processor_examples.get_example_template_builder()
    example_template_builder_2 = processor_examples.get_example_template_builder()
    example_template_builder_3 = processor_examples.get_example_template_builder()

    processor_specification_1 = {
        "region": "r?g",
        "sample": "sig*",
        "systematic": "*",
        "template": "Up",
        "name": "example_template_builder_1",
        "func": example_template_builder_1,
    }
    processor_specification_2 = {
        "region": "abc",
        "sample": "*",
        "systematic": "*",
        "template": "*",
        "name": "example_template_builder_2",
        "func": example_template_builder_2,
    }
    processor_specification_3 = {
        "region": "reg",
        "sample": "bg",
        "systematic": "*",
        "template": None,
        "name": "example_template_builder_3",
        "func": example_template_builder_3,
    }
    example_router.template_builders = [
        processor_specification_1,
        processor_specification_2,
        processor_specification_3,
    ]

    # get a single match
    assert (
        example_router._find_match(
            example_router.template_builders, "reg", "signal", "sys", "Up"
        )
        is example_template_builder_1
    )

    # strings match case 3, but template does not (processor only applied to nominal)
    assert (
        example_router._find_match(
            example_router.template_builders, "reg", "bg", "sys", "Up"
        )
        is None
    )

    # no matches available due to string mismatch
    assert (
        example_router._find_match(
            example_router.template_builders, "reg", "background", "sys", "Up"
        )
        is None
    )

    # match for None (nominal) template with no processor template restrictions ("*")
    assert (
        example_router._find_match(
            example_router.template_builders, "abc", "signal", "sys", None
        )
        is example_template_builder_2
    )

    # match for None (nominal) template with None processor template
    assert (
        example_router._find_match(
            example_router.template_builders, "reg", "bg", "*", None
        )
        is example_template_builder_3
    )

    # multiple matches
    caplog.clear()
    example_router.template_builders = [
        processor_specification_1,
        processor_specification_1,
    ]
    assert (
        example_router._find_match(
            example_router.template_builders, "reg", "signal", "sys", "Up"
        )
        is example_template_builder_1
    )

    assert (
        "found 2 matches, continuing with the first one (example_template_builder)"
        in [rec.message for rec in caplog.records]
    )
    caplog.clear()


def test_Router__find_template_builder_match(processor_examples):
    example_router = route.Router()
    example_template_builder = processor_examples.get_example_template_builder()

    # no wrapper defined
    with pytest.raises(ValueError, match="no template builder wrapper defined"):
        example_router._find_template_builder_match("", "", "", None)

    def example_wrapper(func):
        @functools.wraps(func)
        def wrapper(reg, sam, sys, tem):
            # return the bin yield of the histogram to have something to compare
            return func(reg, sam, sys, tem).values()

        return wrapper

    # wrapper defined, but no match available
    example_router.template_builder_wrapper = example_wrapper
    assert example_router._find_template_builder_match("", "", "", None) is None

    # match exists
    with mock.patch(
        "cabinetry.route.Router._find_match", return_value=example_template_builder
    ) as mock_find:
        wrapped_builder = example_router._find_template_builder_match(
            "reg", "", "", None
        )
        assert mock_find.call_args_list == [(([], "reg", "", "", None), {})]

        # need to verify that wrapped template builder is wrapped with right function
        expected_wrap = example_wrapper(example_template_builder)
        assert wrapped_builder.__name__ == expected_wrap.__name__

        # a direct assert of equality fails
        # assert wrapped_builder == expected_wrap

        # compare instead the behavior for an example call
        assert wrapped_builder({}, {}, {}, {}) == expected_wrap({}, {}, {}, {})


def test_apply_to_all_templates():
    # could mock configuration.histogram_is_needed here
    # define a custom override function that logs its arguments when called
    override_call_args = []

    def match_func(reg: str, sam: str, sys: str, tem: str):
        def f(reg: dict, sam: dict, sys: dict, tem: str):
            override_call_args.append((reg, sam, sys, tem))

        return f

    default_func = mock.MagicMock()

    example_config = {
        "Regions": [{"Name": "test_region"}],
        "Samples": [{"Name": "sample"}],
        "Systematics": [
            {"Name": "norm", "Type": "Normalization"},
            {"Name": "var", "Type": "NormPlusShape"},
        ],
    }

    # no overrides specified
    route.apply_to_all_templates(example_config, default_func, match_func=None)

    # check that the default function was called for all templates
    assert default_func.call_count == 3
    assert default_func.call_args_list[0] == (
        ({"Name": "test_region"}, {"Name": "sample"}, {}, None),
        {},
    )
    assert default_func.call_args_list[1] == (
        (
            {"Name": "test_region"},
            {"Name": "sample"},
            {"Name": "var", "Type": "NormPlusShape"},
            "Up",
        ),
        {},
    )
    assert default_func.call_args_list[2] == (
        (
            {"Name": "test_region"},
            {"Name": "sample"},
            {"Name": "var", "Type": "NormPlusShape"},
            "Down",
        ),
        {},
    )

    # apply the override instead to all templates
    route.apply_to_all_templates(example_config, default_func, match_func=match_func)
    assert len(override_call_args) == 3
    assert override_call_args[0] == (
        {"Name": "test_region"},
        {"Name": "sample"},
        {},
        None,
    )
    assert override_call_args[1] == (
        {"Name": "test_region"},
        {"Name": "sample"},
        {"Name": "var", "Type": "NormPlusShape"},
        "Up",
    )
    assert override_call_args[2] == (
        {"Name": "test_region"},
        {"Name": "sample"},
        {"Name": "var", "Type": "NormPlusShape"},
        "Down",
    )

    # no systematics
    example_config = {
        "Regions": [{"Name": "test_region"}],
        "Samples": [{"Name": "sample"}],
    }
    route.apply_to_all_templates(example_config, default_func)
    # previously 3 calls of default_func, now one more for nominal template
    assert default_func.call_count == 4
    assert default_func.call_args_list[3][0][3] is None



"""Creates required template histograms from columnar data."""

import functools
import logging
import pathlib
from typing import Any, Dict, List, Literal, Optional

import boost_histogram as bh
import numpy as np

from cabinetry import configuration
from cabinetry import histo
from cabinetry import route
from cabinetry.templates import utils


log = logging.getLogger(__name__)


def _ntuple_paths(
    general_path: str,
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> List[pathlib.Path]:
    """Returns the paths to ntuples for a region-sample-systematic-template.

    A path is built starting from the path specified in the general options in the
    configuration file. This path can contain placeholders for region- and sample-
    specific overrides, via ``{Region}`` and ``{Sample}``. For non-nominal templates, it
    is possible to override the path if the ``RegionPath`` or ``SamplePath`` options are
    specified for the template. If ``SamplePath`` is a list, return a list of paths
    (one per entry in the list).

    Args:
        general_path (str): path specified in general settings, with sections that can
            be overridden by region / sample / systematic settings
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Raises:
        ValueError: when ``RegionPath`` placeholder is used, but region setting is not
            specified
        ValueError: when ``SamplePath`` placeholder is used, but sample setting is not
            specified

    Returns:
        List[pathlib.Path]: list of paths to ntuples
    """
    # obtain region and sample paths, if they are defined
    region_path = region.get("RegionPath", None)
    sample_paths = sample.get("SamplePath", None)

    # check whether a systematic is being processed, and whether overrides exist
    if template is not None:
        # determine whether the template has an override for RegionPath specified
        region_override = utils._check_for_override(systematic, template, "RegionPath")
        if region_override is not None:
            region_path = region_override

        # check for SamplePath override
        sample_override = utils._check_for_override(systematic, template, "SamplePath")
        if sample_override is not None:
            sample_paths = sample_override

    region_template_exists = "{RegionPath}" in general_path
    if region_path is not None:
        if not region_template_exists:
            log.warning(
                "region override specified, but {RegionPath} not found in default path"
            )
        general_path = general_path.replace("{RegionPath}", region_path)
    elif region_template_exists:
        raise ValueError(f"no path setting found for region {region['Name']}")

    sample_template_exists = "{SamplePath}" in general_path
    if sample_paths is not None:
        if not sample_template_exists:
            log.warning(
                "sample override specified, but {SamplePath} not found in default path"
            )
        # SamplePath can be a list, so need to construct all possible paths
        sample_paths = configuration._setting_to_list(sample_paths)
        path_list = []
        for sample_path in sample_paths:
            path_list.append(general_path.replace("{SamplePath}", sample_path))
    elif sample_template_exists:
        raise ValueError(f"no path setting found for sample {sample['Name']}")
    else:
        # no need for multiple paths, and no SamplePath are present, so turn
        # the existing path into a list
        path_list = [general_path]

    # convert the contents of path_lists to paths and return them
    paths = [pathlib.Path(path) for path in path_list]
    return paths


def _variable(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> str:
    """Returns the variable the histogram will be binned in.

    For non-nominal templates, overrides the nominal variable if an alternative is
    specified for the template.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Returns:
        str: name of variable to bin histogram in
    """
    axis_variable = region["Variable"]
    # check whether a systematic is being processed
    if template is not None:
        # determine whether the template has an override specified
        axis_variable_override = utils._check_for_override(
            systematic, template, "Variable"
        )
        if axis_variable_override is not None:
            axis_variable = axis_variable_override
    return axis_variable


def _filter(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> Optional[str]:
    """Returns the filter to be applied for event selection.

    Overrides the (optional) filter provided at the region level by a sample-specific
    filter, if provided. For non-nominal templates, overrides the nominal filter if an
    alternative is specified for the template. The systematic-specific filter overrides
    the sample-specific filter if both are provided.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Returns:
        Optional[str]: expression for the filter to be used, or None for no filtering
    """
    selection_filter = region.get("Filter", None)

    # check for sample-specific overrides
    selection_filter_override = sample.get("Filter", None)
    if selection_filter_override is not None:
        selection_filter = selection_filter_override

    # check whether a systematic is being processed
    if template is not None:
        # determine whether the template has an override specified
        selection_filter_override = utils._check_for_override(
            systematic, template, "Filter"
        )
        if selection_filter_override is not None:
            selection_filter = selection_filter_override
    return selection_filter


def _weight(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> Optional[str]:
    """Returns the weight to be used for events in histograms.

    For non-nominal templates, overrides the nominal weight if an alternative is
    specified for the template.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Returns:
        Optional[str]: weight used for events when filled into histograms, or None for
        no weight
    """
    weight = sample.get("Weight", None)
    # check whether a systematic is being processed
    if template is not None:
        # determine whether the template has an override specified
        weight_override = utils._check_for_override(systematic, template, "Weight")
        if weight_override is not None:
            weight = weight_override
    return weight


def _position_in_file(
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> str:
    """Returns the location of data within a file (e.g. a tree name).

    For non-nominal templates, overrides the nominal position if an alternative is
    specified for the template.

    Args:
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Returns:
        str: where in the file to find the data (the name of a tree)
    """
    position = sample["Tree"]
    # check whether a systematic is being processed
    if template is not None:
        # determine whether the template has an override specified
        position_override = utils._check_for_override(systematic, template, "Tree")
        if position_override is not None:
            position = position_override
    return position


def _binning(region: Dict[str, Any]) -> np.ndarray:
    """Returns the binning to be used in a region.

    Should eventually also support other ways of specifying bins, such as the amount of
    bins and the range to bin in.

    Args:
        region (Dict[str, Any]): containing all region information

    Raises:
        NotImplementedError: when the binning is not explicitly defined

    Returns:
        np.ndarray: bin boundaries to be used for histogram
    """
    if not region.get("Binning", False):
        raise NotImplementedError("cannot determine binning")

    return np.asarray(region["Binning"])


class _Builder:
    """Handles the instructions for backends to create histograms."""

    def __init__(
        self, histogram_folder: pathlib.Path, general_path: str, method: str
    ) -> None:
        """Creates an instance, sets histogram folder, path template and method.

        Args:
            histogram_folder (pathlib.Path): folder to save the histograms to
            general_path (str): template for paths to input files for histogram building
            method (str): backend to use for histogram production
        """
        self.histogram_folder = histogram_folder
        self.general_path = general_path
        self.method = method

    def _create_histogram(
        self,
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: Optional[Literal["Up", "Down"]],
    ) -> None:
        """Creates a histogram and writes it to a file.

        The histogram is created for the region-sample-systematic-template specified in
        the argument.

        Args:
            region (Dict[str, Any]): containing all region information
            sample (Dict[str, Any]): containing all sample information
            systematic (Dict[str, Any]): containing all systematic information
            template (Optional[Literal["Up", "Down"]]): template considered: "Up",
                "Down", or None for nominal

        Raises:
            NotImplementedError: when requesting an unknown backend
        """
        ntuple_paths = _ntuple_paths(
            self.general_path, region, sample, systematic, template
        )
        pos_in_file = _position_in_file(sample, systematic, template)
        variable = _variable(region, sample, systematic, template)
        bins = _binning(region)
        weight = _weight(region, sample, systematic, template)
        selection_filter = _filter(region, sample, systematic, template)

        # obtain the histogram
        if self.method == "uproot":
            from cabinetry.contrib import histogram_creator

            histogram = histogram_creator.with_uproot(
                ntuple_paths,
                pos_in_file,
                variable,
                bins,
                weight=weight,
                selection_filter=selection_filter,
            )

        else:
            raise NotImplementedError(f"unknown backend {self.method}")

        # store information in a Histogram instance and save it
        utils._name_and_save(
            self.histogram_folder,
            histo.Histogram(histogram),
            region,
            sample,
            systematic,
            template,
        )

    def _wrap_custom_template_builder(
        self, func: route.UserTemplateFunc
    ) -> route.ProcessorFunc:
        """Returns function that executes custom template builder and saves histogram.

        Wrapper for custom template builder functions that return a
        ``boost_histogram.Histogram``.

        Args:
            func (cabinetry.route.UserTemplateFunc): user-defined template builder

        Returns:
            cabinetry.route.ProcessorFunc: wrapped template builder
        """

        # decorating this with functools.wraps will keep the name of the wrapped
        # function the same, however the signature of the wrapped function is slightly
        # different (the return value becomes None)
        @functools.wraps(func)
        def wrapper(
            region: Dict[str, Any],
            sample: Dict[str, Any],
            systematic: Dict[str, Any],
            template: Optional[Literal["Up", "Down"]],
        ) -> None:
            """Executes a user-defined function that returns a histogram and saves it.

            Returns None, to turn the user-defined function into a ProcessorFunc when
            wrapped with this.

            Args:
                region (Dict[str, Any]): containing all region information
                sample (Dict[str, Any]): containing all sample information
                systematic (Dict[str, Any]): containing all systematic information
                template (Optional[Literal["Up", "Down"]]): template considered: "Up",
                    "Down", or None for nominal
            """
            histogram = func(region, sample, systematic, template)
            if not isinstance(histogram, bh.Histogram):
                raise TypeError(
                    f"{func.__name__} must return a boost_histogram.Histogram"
                )
            utils._name_and_save(
                self.histogram_folder,
                histo.Histogram(histogram),
                region,
                sample,
                systematic,
                template,
            )

        return wrapper



from setuptools import setup

extras_require = {"contrib": ["uproot>=4.1.5"]}  # file writing bug-fixes
extras_require["test"] = sorted(
    set(
        extras_require["contrib"]
        + [
            "pytest",
            "pytest-cov>=2.6.1",  # no_cover support
            "pydocstyle",
            "check-manifest",
            "flake8",
            "flake8-bugbear",
            "flake8-import-order",
            "flake8-print",
            "mypy",
            "types-tabulate",
            "types-PyYAML",
            "typeguard>=4.0.0,!=4.0.1,!=4.1.*",  # cabinetry#391, cabinetry#428
            "black",
        ]
    )
)
extras_require["pyhf_backends"] = ["pyhf[backends]"]
extras_require["docs"] = sorted(
    {
        "sphinx!=5.2.0.post0",  # broken due to version parsing in RTD theme
        "sphinx-click",
        "sphinx-copybutton",
        "sphinx-jsonschema",
        "sphinx-rtd-theme",
    }
)

extras_require["develop"] = sorted(
    set(extras_require["test"] + extras_require["docs"] + ["pre-commit", "twine"])
)
extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

setup(extras_require=extras_require)



"""Constructs HistFactory workspaces in JSON format."""

import json
import logging
import pathlib
from typing import Any, Dict, List, Optional, Union

import pyhf

from cabinetry import configuration
from cabinetry import histo


log = logging.getLogger(__name__)


class WorkspaceBuilder:
    """Collects functionality to build a ``pyhf`` workspace."""

    def __init__(self, config: Dict[str, Any]) -> None:
        """Creates a workspace corresponding to a cabinetry configuration.

        Args:
            config (Dict[str, Any]): ``cabinetry`` configuration
        """
        self.config = config
        self.histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])

    def _data_sample(self) -> Dict[str, Any]:
        """Returns the data sample dictionary.

        Returns:
            Dict[str, Any]: the data sample dictionary
        """
        data_samples = [
            sample for sample in self.config["Samples"] if sample.get("Data", False)
        ]
        if len(data_samples) != 1:
            raise ValueError("did not find exactly one data sample")
        return data_samples[0]

    def _constant_parameter_setting(self, par_name: str) -> Optional[float]:
        """Determines whether a parameter should be set to constant, and to which value.

        For a given parameter, determines if it is supposed to be set to constant. If
        not, returns None, otherwise returns the value it should be fixed to. This only
        looks for the first occurrence of the parameter in the list.

        Args:
            par_name (str): name of parameter to check

        Returns:
            Optional[float]: returns None if parameter is not supposed to be held
            constant, or the value it has to be fixed to
        """
        fixed_parameters = self.config["General"].get("Fixed", [])
        fixed_value = next(
            (
                fixed_par["Value"]
                for fixed_par in fixed_parameters
                if fixed_par["Name"] == par_name
            ),
            None,
        )
        return fixed_value

    def normfactor_modifiers(
        self, region: Dict[str, Any], sample: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Returns the list of NormFactor modifiers acting on a sample in a region.

        Args:
            region (Dict[str, Any]): specific region to get NormFactor modifiers for
            sample (Dict[str, Any]): specific sample to get NormFactor modifiers for

        Returns:
            List[Dict[str, Any]]: NormFactor modifiers for sample
        """
        modifiers = []
        for norm_factor in self.config["NormFactors"]:
            # check that region and sample are both not excluded by modifier
            if configuration.region_contains_modifier(
                region, norm_factor
            ) and configuration.sample_contains_modifier(sample, norm_factor):
                log.debug(
                    f"adding NormFactor {norm_factor['Name']} to sample "
                    f"{sample['Name']} in region {region['Name']}"
                )
                modifiers.append(
                    {"data": None, "name": norm_factor["Name"], "type": "normfactor"}
                )
        return modifiers

    @staticmethod
    def normalization_modifier(systematic: Dict[str, Any]) -> Dict[str, Any]:
        """Returns a normalization modifier (OverallSys in `HistFactory`).

        Args:
            systematic (Dict[str, Any]): systematic for which modifier is constructed

        Returns:
            Dict[str, Any]: single `normsys` modifier for ``pyhf`` workspace
        """
        # take name of modifier from ModifierName if set, default to systematic name
        modifier_name = systematic.get("ModifierName", systematic["Name"])

        modifier = {}
        modifier.update({"name": modifier_name})
        modifier.update({"type": "normsys"})
        modifier.update(
            {
                "data": {
                    "hi": 1 + systematic["Up"]["Normalization"],
                    "lo": 1 + systematic["Down"]["Normalization"],
                }
            }
        )
        return modifier

    def normplusshape_modifiers(
        self, region: Dict[str, Any], sample: Dict[str, Any], systematic: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Returns modifiers for a correlated shape + normalization effect.

        For a variation including a correlated shape + normalization effect, this
        provides the `histosys` and `normsys` modifiers for ``pyhf`` (in `HistFactory`
        language, this corresponds to a `HistoSys` and an `OverallSys`). Symmetrization
        could happen either at this stage (this is the case currently), or somewhere
        earlier, such as during template postprocessing. A `histosys` modifier is not
        created for single-bin channels, as it has no effect in this case (everything is
        handled by the `normsys` modifier already).

        Args:
            region (Dict[str, Any]): region the systematic variation acts in
            sample (Dict[str, Any]): sample the systematic variation acts on
            systematic (Dict[str, Any]): the systematic variation under consideration

        Returns:
            List[Dict[str, Any]]: a list with a ``pyhf`` `normsys` modifier and a
            `histosys` modifier
        """
        # load the systematic variation histogram
        histogram_up = histo.Histogram.from_config(
            self.histogram_folder,
            region,
            sample,
            systematic,
            template="Up",
            modified=True,
        )

        # also need the nominal histogram
        histogram_nominal = histo.Histogram.from_config(
            self.histogram_folder, region, sample, {}, modified=True
        )

        if systematic.get("Down", {}).get("Symmetrize", False):
            # if symmetrization is desired, should support different implementations
            # symmetrization according to "method 1" from issue #26:
            # first normalization, then symmetrization

            # normalize the variation to the same yield as nominal
            norm_effect = histogram_up.normalize_to_yield(histogram_nominal)
            norm_effect_up = norm_effect
            norm_effect_down = 2 - norm_effect
            histo_yield_up = histogram_up.yields.tolist()
            log.debug(
                f"normalization impact of systematic {systematic['Name']} on sample "
                f"{sample['Name']} in region {region['Name']} is {norm_effect:.3f}"
            )
            # need another histogram that corresponds to the "down" variation, which is
            # 2*nominal - up
            histo_yield_down = (
                2 * histogram_nominal.yields - histogram_up.yields
            ).tolist()
        else:
            histogram_down = histo.Histogram.from_config(
                self.histogram_folder,
                region,
                sample,
                systematic,
                template="Down",
                modified=True,
            )
            norm_effect_up = sum(histogram_up.yields) / sum(histogram_nominal.yields)
            norm_effect_down = sum(histogram_down.yields) / sum(
                histogram_nominal.yields
            )
            # normalize templates to same yield as nominal
            histo_yield_up = list(histogram_up.yields / norm_effect_up)
            histo_yield_down = list(histogram_down.yields / norm_effect_down)

        # take name of modifier from ModifierName if set, default to systematic name
        modifier_name = systematic.get("ModifierName", systematic["Name"])

        # add the normsys
        modifiers = []
        norm_modifier = {}
        norm_modifier.update({"name": modifier_name})
        norm_modifier.update({"type": "normsys"})
        norm_modifier.update({"data": {"hi": norm_effect_up, "lo": norm_effect_down}})
        modifiers.append(norm_modifier)

        # add the shape part in a histosys
        if len(histogram_nominal.yields) > 1:
            # only relevant if there is more than one bin, otherwise there is no "shape"
            shape_modifier = {}
            shape_modifier.update({"name": modifier_name})
            shape_modifier.update({"type": "histosys"})
            shape_modifier.update(
                {"data": {"hi_data": histo_yield_up, "lo_data": histo_yield_down}}
            )
            modifiers.append(shape_modifier)
        return modifiers

    def sys_modifiers(
        self, region: Dict[str, Any], sample: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Returns the list of all systematic modifiers acting on a sample in a region.

        Args:
            region (Dict[str, Any]): region considered
            sample (Dict[str, Any]): specific sample to get modifiers for

        Raises:
            NotImplementedError: when unsupported modifiers act on sample

        Returns:
            List[Dict[str, Any]]: modifiers for ``pyhf`` workspace
        """
        modifiers = []
        for systematic in self.config.get("Systematics", []):
            # check that region and sample both do not exclude modifier
            if configuration.region_contains_modifier(
                region, systematic
            ) and configuration.sample_contains_modifier(sample, systematic):
                if systematic["Type"] == "Normalization":
                    # OverallSys (norm uncertainty with Gaussian constraint)
                    log.debug(
                        f"adding OverallSys {systematic['Name']} to sample"
                        f" {sample['Name']} in region {region['Name']}"
                    )
                    modifiers.append(self.normalization_modifier(systematic))
                elif systematic["Type"] == "NormPlusShape":
                    # two modifiers are needed - an OverallSys for the norm effect,
                    # and a HistoSys for the shape variation
                    log.debug(
                        f"adding OverallSys and HistoSys {systematic['Name']} to sample"
                        f" {sample['Name']} in region {region['Name']}"
                    )
                    modifiers += self.normplusshape_modifiers(
                        region, sample, systematic
                    )
                else:
                    raise NotImplementedError(
                        "not supporting other systematic types yet"
                    )
        return modifiers

    def channels(self) -> List[Dict[str, Any]]:
        """Returns the channel information: yields per sample and modifiers.

        Returns:
            List[Dict[str, Any]]: channels for ``pyhf`` workspace
        """
        channels = []
        for region in self.config["Regions"]:
            channel = {}
            channel.update({"name": region["Name"]})
            samples = []
            for sample in self.config["Samples"]:
                if sample.get("Data", False):
                    # skip the data sample, it goes into the observations instead
                    continue

                if not configuration.region_contains_sample(region, sample):
                    # current region does not contain this sample, so skip it
                    continue

                sample_hist = histo.Histogram.from_config(
                    self.histogram_folder, region, sample, {}, modified=True
                )

                # yield of the sample
                current_sample = {}
                current_sample.update({"name": sample["Name"]})
                current_sample.update({"data": sample_hist.yields.tolist()})

                # collect all modifiers for the sample
                modifiers = []

                # gammas
                if not sample.get("DisableStaterror", False):
                    # staterror modifiers are added unless DisableStaterror is True
                    gammas = {}
                    gammas.update(
                        {"name": "staterror_" + region["Name"].replace(" ", "-")}
                    )
                    gammas.update({"type": "staterror"})
                    gammas.update({"data": sample_hist.stdev.tolist()})
                    modifiers.append(gammas)

                # modifiers can have region and sample dependence, which is checked
                # check if normfactors affect sample in region, add modifiers as needed
                nf_modifier_list = self.normfactor_modifiers(region, sample)
                modifiers += nf_modifier_list

                # check if systematics affect sample in region, add modifiers as needed
                sys_modifier_list = self.sys_modifiers(region, sample)
                modifiers += sys_modifier_list

                current_sample.update({"modifiers": modifiers})
                samples.append(current_sample)
            channel.update({"samples": samples})
            channels.append(channel)
        return channels

    def measurements(self) -> List[Dict[str, Any]]:
        """Returns the measurements object for the workspace.

        Constructs the measurements, including POI setting and parameter bounds, initial
        values and whether they are set to constant. Only supports a single measurement
        so far.

        Returns:
            List[Dict[str, Any]]: measurements for ``pyhf`` workspace
        """
        measurements = []
        measurement = {}
        measurement.update({"name": self.config["General"]["Measurement"]})
        config_dict = {}

        # get the norm factor initial values / bounds / constant setting
        parameters_list = []
        for nf in self.config.get("NormFactors", []):
            nf_name = nf["Name"]  # every NormFactor has a name
            init = nf.get("Nominal", None)
            bounds = nf.get("Bounds", None)
            fixed = self._constant_parameter_setting(nf_name)
            if (init is None) and (fixed is not None):
                # if no initial value is specified, but a constant setting
                # is requested, set the initial value to the constant value
                init = fixed

            parameter = {"name": nf_name}
            if init is not None:
                parameter.update({"inits": [init]})
            if bounds is not None:
                parameter.update({"bounds": [bounds]})
            if fixed is not None:
                parameter.update({"fixed": True})

            parameters_list.append(parameter)

        for sys in self.config.get("Systematics", []):
            # when there are many more systematics than NormFactors, it would be more
            # efficient to loop over fixed parameters and exclude all NormFactor related
            # ones to set all the remaining ones to constant (which are systematics)
            sys_name = sys["Name"]  # every systematic has a name
            fixed = self._constant_parameter_setting(sys_name)
            if fixed is not None:
                parameter = {"name": sys_name}
                parameter.update({"inits": [fixed]})
                parameter.update({"fixed": True})
                parameters_list.append(parameter)

        parameters = {"parameters": parameters_list}
        config_dict.update(parameters)
        # POI defaults to "" (interpreted as "no POI" by pyhf) if not specified
        config_dict.update({"poi": self.config["General"].get("POI", "")})
        measurement.update({"config": config_dict})
        measurements.append(measurement)
        return measurements

    def observations(self) -> List[Dict[str, Any]]:
        """Returns the observations object (with data yields) for the workspace.

        Returns:
            List[Dict[str, Any]]: observations for ``pyhf`` workspace
        """
        data_sample = self._data_sample()
        observations = []
        for region in self.config["Regions"]:
            observation = {}
            histo_yield = histo.Histogram.from_config(
                self.histogram_folder, region, data_sample, {}, modified=True
            ).yields.tolist()
            observation.update({"name": region["Name"]})
            observation.update({"data": histo_yield})
            observations.append(observation)
        return observations

    def build(self) -> Dict[str, Any]:
        """Constructs a `HistFactory` workspace in ``pyhf`` format.

        Returns:
            Dict[str, Any]: ``pyhf``-compatible `HistFactory` workspace
        """
        ws: Dict[str, Any] = {}  # the workspace

        # channels
        channels = self.channels()
        ws.update({"channels": channels})

        # measurements
        measurements = self.measurements()
        ws.update({"measurements": measurements})

        # build observations
        observations = self.observations()
        ws.update({"observations": observations})

        # workspace version
        ws.update({"version": "1.0.0"})

        return ws


def build(config: Dict[str, Any], *, with_validation: bool = True) -> Dict[str, Any]:
    """Returns a `HistFactory` workspace in ``pyhf`` format.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        with_validation (bool, optional): validate workspace validity with pyhf,
            defaults to True

    Returns:
        Dict[str, Any]: ``pyhf``-compatible `HistFactory` workspace
    """
    log.info("building workspace")

    ws_builder = WorkspaceBuilder(config)
    ws = ws_builder.build()

    if with_validation:
        validate(ws)
    return ws


def validate(ws: Dict[str, Any]) -> None:
    """Validates a workspace with ``pyhf``.

    Args:
        ws (Dict[str, Any]): the workspace to validate
    """
    pyhf.Workspace(ws)


def save(ws: Dict[str, Any], file_path_string: Union[str, pathlib.Path]) -> None:
    """Serializes a workspace to a file.

    Args:
        ws (Dict[str, Any]): ``pyhf``-compatible `HistFactory` workspace
        file_path_string (Union[str, pathlib.Path]): path to the file to save the
            workspace in
    """
    file_path = pathlib.Path(file_path_string)
    log.debug(f"saving workspace to {file_path}")
    # create output directory if it does not exist yet
    file_path.parent.mkdir(parents=True, exist_ok=True)
    # save workspace to file
    file_path.write_text(json.dumps(ws, sort_keys=True, indent=4))


def load(file_path_string: Union[str, pathlib.Path]) -> Dict[str, Any]:
    """Loads a workspace from a file.

    Args:
        file_path_string (Union[str, pathlib.Path]): path to the file to load the
            workspace from

    Returns:
        Dict[str, Any]: ``pyhf``-compatible `HistFactory` workspace
    """
    file_path = pathlib.Path(file_path_string)
    ws = json.loads(file_path.read_text())
    return ws
